# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Function for generating association rules
#
# Author: Joshua Goerner <https://github.com/JoshuaGoerner>
#         Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from itertools import combinations

import numpy as np
import pandas as pd


def association_rules(df, metric="confidence", min_threshold=0.8, support_only=False):
    """Generates a DataFrame of association rules including the
    metrics 'score', 'confidence', and 'lift'

    Parameters
    -----------
    df : pandas DataFrame
      pandas DataFrame of frequent itemsets
      with columns ['support', 'itemsets']

    metric : string (default: 'confidence')
      Metric to evaluate if a rule is of interest.
      **Automatically set to 'support' if `support_only=True`.**
      Otherwise, supported metrics are 'support', 'confidence', 'lift',
      'leverage', 'conviction' and 'zhangs_metric'
      These metrics are computed as follows:

      - support(A->C) = support(A+C) [aka 'support'], range: [0, 1]\n
      - confidence(A->C) = support(A+C) / support(A), range: [0, 1]\n
      - lift(A->C) = confidence(A->C) / support(C), range: [0, inf]\n
      - leverage(A->C) = support(A->C) - support(A)*support(C),
        range: [-1, 1]\n
      - conviction = [1 - support(C)] / [1 - confidence(A->C)],
        range: [0, inf]\n
      - zhangs_metric(A->C) =
        leverage(A->C) / max(support(A->C)*(1-support(A)), support(A)*(support(C)-support(A->C)))
        range: [-1,1]\n

    min_threshold : float (default: 0.8)
      Minimal threshold for the evaluation metric,
      via the `metric` parameter,
      to decide whether a candidate rule is of interest.

    support_only : bool (default: False)
      Only computes the rule support and fills the other
      metric columns with NaNs. This is useful if:

      a) the input DataFrame is incomplete, e.g., does
      not contain support values for all rule antecedents
      and consequents

      b) you simply want to speed up the computation because
      you don't need the other metrics.

    Returns
    ----------
    pandas DataFrame with columns "antecedents" and "consequents"
      that store itemsets, plus the scoring metric columns:
      "antecedent support", "consequent support",
      "support", "confidence", "lift",
      "leverage", "conviction"
      of all rules for which
      metric(rule) >= min_threshold.
      Each entry in the "antecedents" and "consequents" columns are
      of type `frozenset`, which is a Python built-in type that
      behaves similarly to sets except that it is immutable
      (For more info, see
      https://docs.python.org/3.6/library/stdtypes.html#frozenset).

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/frequent_patterns/association_rules/

    """
    if not df.shape[0]:
        raise ValueError(
            "The input DataFrame `df` containing " "the frequent itemsets is empty."
        )

    # check for mandatory columns
    if not all(col in df.columns for col in ["support", "itemsets"]):
        raise ValueError(
            "Dataframe needs to contain the\
                         columns 'support' and 'itemsets'"
        )

    def conviction_helper(sAC, sA, sC):
        confidence = sAC / sA
        conviction = np.empty(confidence.shape, dtype=float)
        if not len(conviction.shape):
            conviction = conviction[np.newaxis]
            confidence = confidence[np.newaxis]
            sAC = sAC[np.newaxis]
            sA = sA[np.newaxis]
            sC = sC[np.newaxis]
        conviction[:] = np.inf
        conviction[confidence < 1.0] = (1.0 - sC[confidence < 1.0]) / (
            1.0 - confidence[confidence < 1.0]
        )

        return conviction

    def zhangs_metric_helper(sAC, sA, sC):
        denominator = np.maximum(sAC * (1 - sA), sA * (sC - sAC))
        numerator = metric_dict["leverage"](sAC, sA, sC)

        with np.errstate(divide="ignore", invalid="ignore"):
            # ignoring the divide by 0 warning since it is addressed in the below np.where
            zhangs_metric = np.where(denominator == 0, 0, numerator / denominator)

        return zhangs_metric

    # metrics for association rules
    metric_dict = {
        "antecedent support": lambda _, sA, __: sA,
        "consequent support": lambda _, __, sC: sC,
        "support": lambda sAC, _, __: sAC,
        "confidence": lambda sAC, sA, _: sAC / sA,
        "lift": lambda sAC, sA, sC: metric_dict["confidence"](sAC, sA, sC) / sC,
        "leverage": lambda sAC, sA, sC: metric_dict["support"](sAC, sA, sC) - sA * sC,
        "conviction": lambda sAC, sA, sC: conviction_helper(sAC, sA, sC),
        "zhangs_metric": lambda sAC, sA, sC: zhangs_metric_helper(sAC, sA, sC),
    }

    columns_ordered = [
        "antecedent support",
        "consequent support",
        "support",
        "confidence",
        "lift",
        "leverage",
        "conviction",
        "zhangs_metric",
    ]

    # check for metric compliance
    if support_only:
        metric = "support"
    else:
        if metric not in metric_dict.keys():
            raise ValueError(
                "Metric must be 'confidence' or 'lift', got '{}'".format(metric)
            )

    # get dict of {frequent itemset} -> support
    keys = df["itemsets"].values
    values = df["support"].values
    frozenset_vect = np.vectorize(lambda x: frozenset(x))
    frequent_items_dict = dict(zip(frozenset_vect(keys), values))

    # prepare buckets to collect frequent rules
    rule_antecedents = []
    rule_consequents = []
    rule_supports = []

    # iterate over all frequent itemsets
    for k in frequent_items_dict.keys():
        sAC = frequent_items_dict[k]
        # to find all possible combinations
        for idx in range(len(k) - 1, 0, -1):
            # of antecedent and consequent
            for c in combinations(k, r=idx):
                antecedent = frozenset(c)
                consequent = k.difference(antecedent)

                if support_only:
                    # support doesn't need these,
                    # hence, placeholders should suffice
                    sA = None
                    sC = None

                else:
                    try:
                        sA = frequent_items_dict[antecedent]
                        sC = frequent_items_dict[consequent]
                    except KeyError as e:
                        s = (
                            str(e) + "You are likely getting this error"
                            " because the DataFrame is missing "
                            " antecedent and/or consequent "
                            " information."
                            " You can try using the "
                            " `support_only=True` option"
                        )
                        raise KeyError(s)
                    # check for the threshold

                score = metric_dict[metric](sAC, sA, sC)
                if score >= min_threshold:
                    rule_antecedents.append(antecedent)
                    rule_consequents.append(consequent)
                    rule_supports.append([sAC, sA, sC])

    # check if frequent rule was generated
    if not rule_supports:
        return pd.DataFrame(columns=["antecedents", "consequents"] + columns_ordered)

    else:
        # generate metrics
        rule_supports = np.array(rule_supports).T.astype(float)
        df_res = pd.DataFrame(
            data=list(zip(rule_antecedents, rule_consequents)),
            columns=["antecedents", "consequents"],
        )

        if support_only:
            sAC = rule_supports[0]
            for m in columns_ordered:
                df_res[m] = np.nan
            df_res["support"] = sAC

        else:
            sAC = rule_supports[0]
            sA = rule_supports[1]
            sC = rule_supports[2]
            for m in columns_ordered:
                df_res[m] = metric_dict[m](sAC, sA, sC)

        return df_res



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Bootstrap functions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from itertools import product

import numpy as np
from sklearn.base import clone

from .bootstrap_outofbag import BootstrapOutOfBag


def _check_arrays(X, y=None):
    if isinstance(X, list):
        raise ValueError("X must be a numpy array")
    if not len(X.shape) == 2:
        raise ValueError("X must be a 2D array. Try X[:, numpy.newaxis]")
    try:
        if y is None:
            return
    except AttributeError:
        if not len(y.shape) == 1:
            raise ValueError("y must be a 1D array.")

    if not len(y) == X.shape[0]:
        raise ValueError("X and y must contain the" "same number of samples")


def no_information_rate(targets, predictions, loss_fn):
    combinations = np.array(list(product(targets, predictions)))
    return loss_fn(combinations[:, 0], combinations[:, 1])


def accuracy(targets, predictions):
    return np.mean(np.array(targets) == np.array(predictions))


def mse(targets, predictions):
    return np.mean((np.array(targets) - np.array(predictions)) ** 2)


def bootstrap_point632_score(
    estimator,
    X,
    y,
    n_splits=200,
    method=".632",
    scoring_func=None,
    predict_proba=False,
    random_seed=None,
    clone_estimator=True,
    **fit_params,
):
    """
    Implementation of the .632 [1] and .632+ [2] bootstrap
    for supervised learning

    References:

    - [1] Efron, Bradley. 1983. "Estimating the Error Rate
      of a Prediction Rule: Improvement on Cross-Validation."
      Journal of the American Statistical Association
      78 (382): 316. doi:10.2307/2288636.
    - [2] Efron, Bradley, and Robert Tibshirani. 1997.
      "Improvements on Cross-Validation: The .632+ Bootstrap Method."
      Journal of the American Statistical Association
      92 (438): 548. doi:10.2307/2965703.

    Parameters
    ----------
    estimator : object
        An estimator for classification or regression that
        follows the scikit-learn API and implements "fit" and "predict"
        methods.

    X : array-like
        The data to fit. Can be, for example a list, or an array at least 2d.

    y : array-like, optional, default: None
        The target variable to try to predict in the case of
        supervised learning.

    n_splits : int (default=200)
        Number of bootstrap iterations.
        Must be larger than 1.

    method : str (default='.632')
        The bootstrap method, which can be either
        - 1) '.632' bootstrap (default)
        - 2) '.632+' bootstrap
        - 3) 'oob' (regular out-of-bag, no weighting)
        for comparison studies.

    scoring_func : callable,
        Score function (or loss function) with signature
        ``scoring_func(y, y_pred, **kwargs)``.
        If none, uses classification accuracy if the
        estimator is a classifier and mean squared error
        if the estimator is a regressor.

    predict_proba : bool
        Whether to use the `predict_proba` function for the
        `estimator` argument. This is to be used in conjunction
        with `scoring_func` which takes in probability values
        instead of actual predictions.
        For example, if the scoring_func is
        :meth:`sklearn.metrics.roc_auc_score`, then use
        `predict_proba=True`.
        Note that this requires `estimator` to have
        `predict_proba` method implemented.

    random_seed : int (default=None)
        If int, random_seed is the seed used by
        the random number generator.

    clone_estimator : bool (default=True)
        Clones the estimator if true, otherwise fits
        the original.

    fit_params : additional parameters
        Additional parameters to be passed to the .fit() function of the
        estimator when it is fit to the bootstrap samples.


    Returns
    -------
    scores : array of float, shape=(len(list(n_splits)),)
        Array of scores of the estimator for each bootstrap
        replicate.

    Examples
    --------
    >>> from sklearn import datasets, linear_model
    >>> from mlxtend.evaluate import bootstrap_point632_score
    >>> iris = datasets.load_iris()
    >>> X = iris.data
    >>> y = iris.target
    >>> lr = linear_model.LogisticRegression()
    >>> scores = bootstrap_point632_score(lr, X, y)
    >>> acc = np.mean(scores)
    >>> print('Accuracy:', acc)
    0.953023146884
    >>> lower = np.percentile(scores, 2.5)
    >>> upper = np.percentile(scores, 97.5)
    >>> print('95%% Confidence interval: [%.2f, %.2f]' % (lower, upper))
    95% Confidence interval: [0.90, 0.98]

    For more usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/bootstrap_point632_score/

    """
    if not isinstance(n_splits, int) or n_splits < 1:
        raise ValueError(
            "Number of splits must be" " greater than 1. Got %s." % n_splits
        )

    allowed_methods = (".632", ".632+", "oob")
    if not isinstance(method, str) or method not in allowed_methods:
        raise ValueError(
            "The `method` must " "be in %s. Got %s." % (allowed_methods, method)
        )

    # Pandas compatibility
    if hasattr(X, "values"):
        X = X.values
    if hasattr(y, "values"):
        y = y.values

    _check_arrays(X, y)

    if clone_estimator:
        cloned_est = clone(estimator)
    else:
        cloned_est = estimator

    if scoring_func is None:
        if cloned_est._estimator_type == "classifier":
            scoring_func = accuracy
        elif cloned_est._estimator_type == "regressor":
            scoring_func = mse
        else:
            raise AttributeError(
                "Estimator type undefined." "Please provide a scoring_func argument."
            )

    # determine which prediction function to use
    # either label, or probability prediction
    if not predict_proba:
        predict_func = cloned_est.predict
    else:
        if not getattr(cloned_est, "predict_proba", None):
            raise RuntimeError(
                f"The estimator {cloned_est} does not "
                f"support predicting probabilities via "
                f"`predict_proba` function."
            )
        predict_func = cloned_est.predict_proba

    oob = BootstrapOutOfBag(n_splits=n_splits, random_seed=random_seed)
    scores = np.empty(dtype=float, shape=(n_splits,))
    cnt = 0

    for train, test in oob.split(X):
        cloned_est.fit(X[train], y[train], **fit_params)

        # get the prediction probability
        # for binary class uses the last column
        predicted_test_val = predict_func(X[test])

        if method in (".632", ".632+"):
            # predictions on the internal training set:
            # predicted_train_val = predict_func(X[train])

            # compute training error on the whole training set as reported in
            # the original .632 boostrap paper
            # in Eq (6.12) in
            #    "Estimating the Error Rate of a Prediction Rule: Improvement
            #     on Cross-Validation"
            #     by B. Efron, 1983, https://doi.org/10.2307/2288636
            # Also see the discussion at
            #   https://github.com/rasbt/mlxtend/discussions/828
            #
            # This also applies to the .632+ estimate in the paper
            #    "Improvements on Cross-Validation: The .632+ Bootstrap Method"
            #    https://www.tandfonline.com/doi/abs/10.1080/01621459.1997.10474007
            predicted_train_val = predict_func(X)

        if predict_proba:
            len_uniq = np.unique(y)

            if len(len_uniq) == 2:
                predicted_train_val = predicted_train_val[:, 1]
                predicted_test_val = predicted_test_val[:, 1]

        test_acc = scoring_func(y[test], predicted_test_val)

        if method == "oob":
            acc = test_acc

        else:
            test_err = 1 - test_acc

            # training error on the whole training set as mentioned in the
            # previous comment above
            train_err = 1 - scoring_func(y, predicted_train_val)

            if method == ".632+":
                gamma = 1 - (
                    no_information_rate(y, cloned_est.predict(X), scoring_func)
                )
                R = (test_err - train_err) / (gamma - train_err)
                weight = 0.632 / (1 - 0.368 * R)

            else:
                weight = 0.632

            acc = 1 - (weight * test_err + (1.0 - weight) * train_err)

        scores[cnt] = acc
        cnt += 1
    return scores



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Algorithm for sequential feature selection.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import datetime
import sys
import types
from copy import deepcopy
from itertools import combinations

import numpy as np
import scipy as sp
import scipy.stats
from joblib import Parallel, delayed
from sklearn.base import MetaEstimatorMixin, clone
from sklearn.metrics import get_scorer

from ..externals.name_estimators import _name_estimators
from ..utils.base_compostion import _BaseXComposition
from .utilities import _calc_score, _get_featurenames, _merge_lists, _preprocess


class SequentialFeatureSelector(_BaseXComposition, MetaEstimatorMixin):

    """Sequential Feature Selection for Classification and Regression.

    Parameters
    ----------
    estimator : scikit-learn classifier or regressor
    k_features : int or tuple or str (default: 1)
        Number of features to select,
        where k_features < the full feature set.
        New in 0.4.2: A tuple containing a min and max value can be provided,
            and the SFS will consider return any feature combination between
            min and max that scored highest in cross-validation. For example,
            the tuple (1, 4) will return any combination from
            1 up to 4 features instead of a fixed number of features k.
        New in 0.8.0: A string argument "best" or "parsimonious".
            If "best" is provided, the feature selector will return the
            feature subset with the best cross-validation performance.
            If "parsimonious" is provided as an argument, the smallest
            feature subset that is within one standard error of the
            cross-validation performance will be selected.

    forward : bool (default: True)
        Forward selection if True,
        backward selection otherwise

    floating : bool (default: False)
        Adds a conditional exclusion/inclusion if True.

    verbose : int (default: 0), level of verbosity to use in logging.
        If 0, no output,
        if 1 number of features in current set, if 2 detailed logging i
        ncluding timestamp and cv scores at step.

    scoring : str, callable, or None (default: None)
        If None (default), uses 'accuracy' for sklearn classifiers
        and 'r2' for sklearn regressors.
        If str, uses a sklearn scoring metric string identifier, for example
        {accuracy, f1, precision, recall, roc_auc} for classifiers,
        {'mean_absolute_error', 'mean_squared_error'/'neg_mean_squared_error',
        'median_absolute_error', 'r2'} for regressors.
        If a callable object or function is provided, it has to be conform with
        sklearn's signature ``scorer(estimator, X, y)``; see
        https://scikit-learn.org/stable/modules/generated/sklearn.metrics.make_scorer.html
        for more information.

    cv : int (default: 5)
        Integer or iterable yielding train, test splits. If cv is an integer
        and `estimator` is a classifier (or y consists of integer class
        labels) stratified k-fold. Otherwise regular k-fold cross-validation
        is performed. No cross-validation if cv is None, False, or 0.

    n_jobs : int (default: 1)
        The number of CPUs to use for evaluating different feature subsets
        in parallel. -1 means 'all CPUs'.

    pre_dispatch : int, or string (default: '2*n_jobs')
        Controls the number of jobs that get dispatched
        during parallel execution if `n_jobs > 1` or `n_jobs=-1`.
        Reducing this number can be useful to avoid an explosion of
        memory consumption when more jobs get dispatched than CPUs can process.
        This parameter can be:
        None, in which case all the jobs are immediately created and spawned.
            Use this for lightweight and fast-running jobs,
            to avoid delays due to on-demand spawning of the jobs
        An int, giving the exact number of total jobs that are spawned
        A string, giving an expression as a function
            of n_jobs, as in `2*n_jobs`

    clone_estimator : bool (default: True)
        Clones estimator if True; works with the original estimator instance
        if False. Set to False if the estimator doesn't
        implement scikit-learn's set_params and get_params methods.
        In addition, it is required to set cv=0, and n_jobs=1.

    fixed_features : tuple (default: None)
        If not `None`, the feature indices provided as a tuple will be
        regarded as fixed by the feature selector. For example, if
        `fixed_features=(1, 3, 7)`, the 2nd, 4th, and 8th feature are
        guaranteed to be present in the solution. Note that if
        `fixed_features` is not `None`, make sure that the number of
        features to be selected is greater than `len(fixed_features)`.
        In other words, ensure that `k_features > len(fixed_features)`.
        New in mlxtend v. 0.18.0.

    feature_groups : list or None (default: None)
        Optional argument for treating certain features as a group.
        This means, the features within a group are always selected together,
        never split.
        For example, `feature_groups=[[1], [2], [3, 4, 5]]`
        specifies 3 feature groups. In this case,
        possible feature selection results with `k_features=2`
        are `[[1], [2]`, `[[1], [3, 4, 5]]`, or `[[2], [3, 4, 5]]`.
        Feature groups can be useful for
        interpretability, for example, if features 3, 4, 5 are one-hot
        encoded features.  (For  more details, please read the notes at the
        bottom of this docstring).  New in mlxtend v. 0.21.0.

    Attributes
    ----------
    k_feature_idx_ : array-like, shape = [n_predictions]
        Feature Indices of the selected feature subsets.

    k_feature_names_ : array-like, shape = [n_predictions]
        Feature names of the selected feature subsets. If pandas
        DataFrames are used in the `fit` method, the feature
        names correspond to the column names. Otherwise, the
        feature names are string representation of the feature
        array indices. New in v 0.13.0.

    k_score_ : float
        Cross validation average score of the selected subset.

    subsets_ : dict
        A dictionary of selected feature subsets during the
        sequential selection, where the dictionary keys are
        the lengths k of these feature subsets. If the parameter
        `feature_groups` is not None, the value of key indicates
        the number of groups that are selected together. The dictionary
        values are dictionaries themselves with the following
        keys: 'feature_idx' (tuple of indices of the feature subset)
              'feature_names' (tuple of feature names of the feat. subset)
              'cv_scores' (list individual cross-validation scores)
              'avg_score' (average cross-validation score)
        Note that if pandas
        DataFrames are used in the `fit` method, the 'feature_names'
        correspond to the column names. Otherwise, the
        feature names are string representation of the feature
        array indices. The 'feature_names' is new in v 0.13.0.

    Notes
    -----
    (1) If parameter `feature_groups` is not None, the
    number of features is equal to the number of feature groups, i.e.
    `len(feature_groups)`. For  example, if `feature_groups = [[0], [1], [2, 3],
    [4]]`, then the `max_features` value cannot exceed 4.

    (2) Although two or more individual features may be considered as one group
    throughout the feature-selection process, it does not mean the individual
    features of that group have the same impact on the outcome. For instance, in
    linear regression, the coefficient of the feature 2 and 3 can be different
    even if they are considered as one group in feature_groups.

    (3) If both fixed_features and feature_groups are specified, ensure that each
    feature group contains the fixed_features selection. E.g., for a 3-feature set
    fixed_features=[0, 1] and feature_groups=[[0, 1], [2]] is valid;
    fixed_features=[0, 1] and feature_groups=[[0], [1, 2]] is not valid.

    (4) In case of KeyboardInterrupt, the dictionary subsets may not be completed.
    If user is still interested in getting the best score, they can use method
    `finalize_fit`.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/feature_selection/SequentialFeatureSelector/

    """

    def __init__(
        self,
        estimator,
        k_features=1,
        forward=True,
        floating=False,
        verbose=0,
        scoring=None,
        cv=5,
        n_jobs=1,
        pre_dispatch="2*n_jobs",
        clone_estimator=True,
        fixed_features=None,
        feature_groups=None,
    ):
        self.estimator = estimator
        self.k_features = k_features
        self.forward = forward
        self.floating = floating
        self.pre_dispatch = pre_dispatch
        # Want to raise meaningful error message if a
        # cross-validation generator is inputted
        if isinstance(cv, types.GeneratorType):
            err_msg = (
                "Input cv is a generator object, which is not "
                "supported. Instead please input an iterable yielding "
                "train, test splits. This can usually be done by "
                "passing a cross-validation generator to the "
                "built-in list function. I.e. cv=list(<cv-generator>)"
            )
            raise TypeError(err_msg)
        self.cv = cv
        self.n_jobs = n_jobs
        self.verbose = verbose

        self.clone_estimator = clone_estimator
        if self.clone_estimator:
            self.est_ = clone(self.estimator)
        else:
            self.est_ = self.estimator

        self.scoring = scoring
        if self.scoring is None:
            if not hasattr(self.est_, "_estimator_type"):
                raise AttributeError(
                    "Estimator must have an ._estimator_type for infering `scoring`"
                )

            if self.est_._estimator_type == "classifier":
                self.scoring = "accuracy"
            elif self.est_._estimator_type == "regressor":
                self.scoring = "r2"
            else:
                raise AttributeError("Estimator must be a Classifier or Regressor.")

        if isinstance(self.scoring, str):
            self.scorer = get_scorer(self.scoring)
        else:
            self.scorer = self.scoring

        self.fixed_features = fixed_features
        self.feature_groups = feature_groups

        self.fitted = False
        self.interrupted_ = False

        # don't mess with this unless testing
        self._TESTING_INTERRUPT_MODE = False

    @property
    def named_estimators(self):
        """
        Returns
        -------
        List of named estimator tuples, like [('svc', SVC(...))]
        """
        return _name_estimators([self.estimator])

    def get_params(self, deep=True):
        #
        # Return estimator parameter names for GridSearch support.
        #
        return self._get_params("named_estimators", deep=deep)

    def set_params(self, **params):
        """Set the parameters of this estimator.
        Valid parameter keys can be listed with ``get_params()``.

        Returns
        -------
        self
        """
        self._set_params("estimator", "named_estimators", **params)
        return self

    def generate_error_message_k_features(self, name):
        if (
            len(self.fixed_features_) == 0
            and len(self.feature_groups_) == self.n_features
        ):
            err_msg = f"{name} must be between 1 and X.shape[1]."

        elif (
            len(self.fixed_features_) > 0
            and len(self.feature_groups_) == self.n_features
        ):
            err_msg = f"{name} must be between len(fixed_features) and X.shape[1]."

        elif (
            len(self.fixed_features_) == 0
            and len(self.feature_groups_) < self.n_features
        ):
            err_msg = f"{name} must be between 1 and len(feature_groups)."

        else:  # both fixed_features and feature_groups are provided
            err_msg = f"{name} must be between the number of groups that appear in fixed_features and len(feature_groups)."

        return err_msg

    def fit(self, X, y, groups=None, **fit_params):
        """Perform feature selection and learn model from training data.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.
            New in v 0.13.0: pandas DataFrames are now also accepted as
            argument for X.
        y : array-like, shape = [n_samples]
            Target values.
            New in v 0.13.0: pandas DataFrames are now also accepted as
            argument for y.
        groups : array-like, with shape (n_samples,), optional
            Group labels for the samples used while splitting the dataset into
            train/test set. Passed to the fit method of the cross-validator.
        fit_params : various, optional
            Additional parameters that are being passed to the estimator.
            For example, `sample_weights=weights`.

        Returns
        -------
        self : object

        """

        # reset from a potential previous fit run
        self.subsets_ = {}
        self.fitted = False
        self.interrupted_ = False
        self.k_feature_idx_ = None
        self.k_feature_names_ = None
        self.k_score_ = None

        X_, self.feature_names = _preprocess(X)
        self.n_features = X_.shape[1]

        self.feature_names_to_idx_mapper = None
        if self.feature_names is not None:
            self.feature_names_to_idx_mapper = {
                name: idx for idx, name in enumerate(self.feature_names)
            }

        self.fixed_features_ = self.fixed_features
        if self.fixed_features_ is None:
            self.fixed_features_ = tuple()

        fixed_feature_types = {type(i) for i in self.fixed_features_}
        if len(fixed_feature_types) > 1:
            raise ValueError(
                f"fixed_features values must have the same type. Found {fixed_feature_types}."
            )

        if len(self.fixed_features_) > 0 and isinstance(self.fixed_features_[0], str):
            if self.feature_names_to_idx_mapper is None:
                raise ValueError(
                    "The input X does not contain name of features provived in"
                    " `fixed_features`. Try passing input X as pandas DataFrames."
                )

            self.fixed_features_ = tuple(
                self.feature_names_to_idx_mapper[name] for name in self.fixed_features_
            )

        if not set(self.fixed_features_).issubset(set(range(X_.shape[1]))):
            raise ValueError(
                "`fixed_features` contains at least one feature that is not in the"
                " input data `X`."
            )

        self.feature_groups_ = self.feature_groups
        if self.feature_groups_ is None:
            self.feature_groups_ = [[i] for i in range(X_.shape[1])]

        for fg in self.feature_groups_:
            if len(fg) == 0:
                raise ValueError(
                    "Each list in the nested lists `features_group` cannot be empty"
                )

        feature_group_types = {
            type(i) for sublist in self.feature_groups_ for i in sublist
        }
        if len(feature_group_types) > 1:
            raise ValueError(
                f"feature_group values must have the same type. Found {feature_group_types}."
            )

        if isinstance(self.feature_groups_[0][0], str):
            if self.feature_names_to_idx_mapper is None:
                raise ValueError(
                    "The input X does not contain name of features provived in"
                    " `feature_groups`. Try passing input X as pandas DataFrames"
                    " in which the name of features match the ones provided in"
                    " `feature_groups`"
                )

            lst = []
            for item in self.feature_groups_:
                tmp = [self.feature_names_to_idx_mapper[name] for name in item]
                lst.append(tmp)

            self.feature_groups_ = lst

        if sorted(_merge_lists(self.feature_groups_)) != sorted(
            list(range(X_.shape[1]))
        ):
            raise ValueError(
                "`feature_group` must contain all features within `range(X.shape[1])`"
                " and there should be no common feature betweeen any two distinct"
                " group of features provided in `feature_group`"
            )

        features_encoded_by_groupID = np.full(X_.shape[1], -1, dtype=np.int64)
        for group_id, group in enumerate(self.feature_groups_):
            for idx in group:
                features_encoded_by_groupID[idx] = group_id

        lst = [features_encoded_by_groupID[idx] for idx in self.fixed_features_]
        self.fixed_features_group_set = set(lst)

        n_fixed_features_expected = sum(
            len(self.feature_groups_[group_id])
            for group_id in self.fixed_features_group_set
        )
        if n_fixed_features_expected != len(self.fixed_features_):
            raise ValueError(
                "For each feature specified in the `fixed feature`, its group-mates"
                "must be specified as `fix_features` as well when `feature_groups`"
                "is provided."
            )

        self.k_lb = max(1, len(self.fixed_features_group_set))
        self.k_ub = len(self.feature_groups_)

        if (
            not isinstance(self.k_features, int)
            and not isinstance(self.k_features, tuple)
            and not isinstance(self.k_features, str)
        ):
            raise AttributeError(
                "k_features must be a positive integer" ", tuple, or string"
            )

        eligible_k_values_range = range(self.k_lb, self.k_ub + 1)
        if isinstance(self.k_features, int) and (
            self.k_features not in eligible_k_values_range
        ):
            err_msg = self.generate_error_message_k_features("k_features")
            raise AttributeError(err_msg)

        if isinstance(self.k_features, tuple):
            if len(self.k_features) != 2:
                raise AttributeError(
                    "k_features tuple must consist of 2"
                    " elements, a min and a max value."
                )

            if self.k_features[0] > self.k_features[1]:
                raise AttributeError(
                    "The min k_features value must be smaller"
                    " than the max k_features value."
                )

            if self.k_features[0] not in eligible_k_values_range:
                err_msg = self.generate_error_message_k_features(
                    "k_features tuple min value"
                )
                raise AttributeError(err_msg)

                # raise AttributeError(
                #    "k_features tuple min value must be in" " range(1, X.shape[1]+1)."
                # )

            if self.k_features[1] not in eligible_k_values_range:
                err_msg = self.generate_error_message_k_features(
                    "k_features tuple max value"
                )
                raise AttributeError(err_msg)

                # raise AttributeError(
                #    "k_features tuple max value must be in" " range(1, X.shape[1]+1)."
                # )

        self.is_parsimonious = False
        if isinstance(self.k_features, str):
            if self.k_features not in {"best", "parsimonious"}:
                raise AttributeError(
                    "If a string argument is provided, "
                    'it must be "best" or "parsimonious"'
                )
            if self.k_features == "parsimonious":
                self.is_parsimonious = True

        if isinstance(self.k_features, str):
            self.k_features = (self.k_lb, self.k_ub)
        elif isinstance(self.k_features, int):
            # we treat k_features as k group of features
            self.k_features = (self.k_features, self.k_features)

        self.min_k = self.k_features[0]
        self.max_k = self.k_features[1]

        if self.forward:
            k_idx = tuple(sorted(self.fixed_features_group_set))
            k_stop = self.max_k
        else:
            k_idx = tuple(range(self.k_ub))
            k_stop = self.min_k

        k = len(k_idx)
        if k > 0:
            k_idx, k_score = _calc_score(
                self,
                X_,
                y,
                k_idx,
                groups=groups,
                feature_groups=self.feature_groups_,
                **fit_params,
            )
            self.subsets_[k] = {
                "feature_idx": k_idx,
                "cv_scores": k_score,
                "avg_score": np.nanmean(k_score),
            }

        orig_set = set(range(self.k_ub))
        try:
            while k != k_stop:
                prev_subset = set(k_idx)
                if self.forward:
                    search_set = orig_set
                    must_include_set = prev_subset
                else:
                    search_set = prev_subset
                    must_include_set = self.fixed_features_group_set

                k_idx, k_score, cv_scores = self._feature_selector(
                    search_set,
                    must_include_set,
                    X=X_,
                    y=y,
                    is_forward=self.forward,
                    groups=groups,
                    feature_groups=self.feature_groups_,
                    **fit_params,
                )

                k = len(k_idx)
                # floating can lead to multiple same-sized subsets
                if k not in self.subsets_ or (k_score > self.subsets_[k]["avg_score"]):
                    k_idx = tuple(sorted(k_idx))
                    self.subsets_[k] = {
                        "feature_idx": k_idx,
                        "cv_scores": cv_scores,
                        "avg_score": k_score,
                    }

                if self.floating:
                    # floating direction is opposite of self.forward, i.e. in
                    # forward selection, we do floating in backward manner,
                    # and in backward selection, we do floating in forward manner
                    is_float_forward = not self.forward
                    (new_feature_idx,) = set(k_idx) ^ prev_subset
                    for _ in range(X_.shape[1]):
                        if (
                            self.forward
                            and (len(k_idx) - len(self.fixed_features_group_set)) <= 2
                        ):
                            break
                        if not self.forward and (len(orig_set) - len(k_idx) <= 2):
                            break

                        if is_float_forward:
                            # corresponding to self.forward=False
                            search_set = orig_set - {new_feature_idx}
                            must_include_set = set(k_idx)
                        else:
                            # corresponding to self.forward=True
                            search_set = set(k_idx)
                            must_include_set = self.fixed_features_group_set | {
                                new_feature_idx
                            }

                        (
                            k_idx_c,
                            k_score_c,
                            cv_scores_c,
                        ) = self._feature_selector(
                            search_set,
                            must_include_set,
                            X=X_,
                            y=y,
                            is_forward=is_float_forward,
                            groups=groups,
                            feature_groups=self.feature_groups_,
                            **fit_params,
                        )

                        if k_score_c <= k_score:
                            break

                        # In the floating process, we basically revisit our previous
                        # steps. so, len(k_idx_c) definitely exists as a key in
                        # the dictionary `self.subsets_`
                        if k_score_c <= self.subsets_[len(k_idx_c)]["avg_score"]:
                            break
                        else:
                            k_idx, k_score, cv_scores = k_idx_c, k_score_c, cv_scores_c
                            k_idx = tuple(sorted(k_idx))
                            k = len(k_idx)
                            self.subsets_[k] = {
                                "feature_idx": k_idx,
                                "cv_scores": cv_scores,
                                "avg_score": k_score,
                            }

                if self.verbose == 1:
                    sys.stderr.write("\rFeatures: %d/%s" % (len(k_idx), k_stop))
                    sys.stderr.flush()
                elif self.verbose > 1:
                    sys.stderr.write(
                        "\n[%s] Features: %d/%s -- score: %s"
                        % (
                            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                            len(k_idx),
                            k_stop,
                            k_score,
                        )
                    )

                if self._TESTING_INTERRUPT_MODE:  # just to test `KeyboardInterrupt`
                    self.finalize_fit()
                    raise KeyboardInterrupt

        except KeyboardInterrupt:
            self.interrupted_ = True
            sys.stderr.write("\nSTOPPING EARLY DUE TO KEYBOARD INTERRUPT...")

        if self.interrupted_:
            self.fitted = False
        else:
            self.fitted = True  # the completion of sequential selection process.
            self.finalize_fit()

        return self

    def finalize_fit(self):
        max_score = np.NINF
        for k in self.subsets_:
            if (
                k >= self.min_k
                and k <= self.max_k
                and self.subsets_[k]["avg_score"] > max_score
            ):
                max_score = self.subsets_[k]["avg_score"]
                best_subset = k

        k_score = max_score
        if k_score == np.NINF:
            # i.e. all keys of self.subsets_ are not in interval `[self.min_k, self.max_k]`
            # this happens if KeyboardInterrupt happens
            keys = list(self.subsets_.keys())
            scores = [self.subsets_[k]["avg_score"] for k in keys]
            arg = np.argmax(scores)

            k_score = scores[arg]
            best_subset = keys[arg]

        k_idx = self.subsets_[best_subset]["feature_idx"]

        if self.is_parsimonious:
            for k in self.subsets_:
                if k >= best_subset:
                    continue
                if self.subsets_[k]["avg_score"] >= (
                    max_score
                    - np.std(self.subsets_[k]["cv_scores"])
                    / self.subsets_[k]["cv_scores"].shape[0]
                ):
                    max_score = self.subsets_[k]["avg_score"]
                    best_subset = k

            k_score = max_score
            k_idx = self.subsets_[best_subset]["feature_idx"]

        for k in self.subsets_:
            self.subsets_[k]["feature_idx"] = _merge_lists(
                self.feature_groups_, self.subsets_[k]["feature_idx"]
            )
        self.k_feature_idx_ = _merge_lists(self.feature_groups_, k_idx)
        self.k_score_ = k_score
        self.subsets_, self.k_feature_names_ = _get_featurenames(
            self.subsets_, self.k_feature_idx_, self.feature_names, self.n_features
        )

        return

    def _feature_selector(
        self,
        search_set,
        must_include_set,
        X,
        y,
        is_forward,
        groups=None,
        feature_groups=None,
        **fit_params,
    ):
        """Perform one round of feature selection. When `is_forward=True`, it is
        a forward selection that searches the `search_set` to find one feature that
        with `must_include_set` results in highest average score. When
        `is_forward=False`, it is a backward selection that searches the `search_set`
        for a feature that its exclusion results in a set of features that includes
        `must_include_set` and has the highest averege score.

        Parameters
        ----------
        self : object
            an instance of class `SequentialFeatureSelector`

        search_set : set
            a set of features through which a feature must be selected to be included
            (when `is_forward=True`) or to be excluded (when `is_forward=False`)

        must_include_set : set
            a set of features that must be present in the selected subset of features

        X : numpy.ndarray
            a 2D numpy array. Each row corresponds to one observation and each
            column corresponds to one feature.

        y : numpy.ndarray
            the target variable

        is_forward : bool
            True if it is forward selection. False if it is backward selection

        groups : array-like, with shape (n_samples,), optional
            Group labels for the samples used while splitting the dataset into
            train/test set. Passed to the fit method of the cross-validator.

        feature_groups : list or None (default: None)
            Optional argument for treating certain features as a group.

        fit_params : various, optional
            Additional parameters that are being passed to the estimator.
            For example, `sample_weights=weights`.

        Returns
        -------
        out1 : the selected set of features that has the highest mean of cv scores
        out2 : the mean of cv scores for the selected set of features.
        out3 : all cv scores for the selected set of features
        """
        out = (None, None, None)

        if feature_groups is None:
            feature_groups = [[i] for i in range(X.shape[1])]

        remaining_set = search_set - must_include_set
        remaining = list(remaining_set)
        n = len(remaining)
        if n > 0:
            if is_forward:
                feature_explorer = combinations(remaining, r=1)
            else:
                feature_explorer = combinations(remaining, r=n - 1)

            n_jobs = min(self.n_jobs, n)
            parallel = Parallel(
                n_jobs=n_jobs, verbose=self.verbose, pre_dispatch=self.pre_dispatch
            )
            work = parallel(
                delayed(_calc_score)(
                    self,
                    X,
                    y,
                    tuple(set(p) | must_include_set),
                    groups=groups,
                    feature_groups=feature_groups,
                    **fit_params,
                )
                for p in feature_explorer
            )

            all_avg_scores = []
            all_cv_scores = []
            all_subsets = []
            for new_subset, cv_scores in work:
                all_avg_scores.append(np.nanmean(cv_scores))
                all_cv_scores.append(cv_scores)
                all_subsets.append(new_subset)

            if len(all_avg_scores) > 0:
                best = np.argmax(all_avg_scores)
                out = (all_subsets[best], all_avg_scores[best], all_cv_scores[best])

        return out

    def transform(self, X):
        """Reduce X to its most important features.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.
            New in v 0.13.0: pandas DataFrames are now also accepted as
            argument for X.

        Returns
        -------
        Reduced feature subset of X, shape={n_samples, k_features}

        """
        self._check_fitted()
        X_, _ = _preprocess(X)
        return X_[:, self.k_feature_idx_]

    def fit_transform(self, X, y, groups=None, **fit_params):
        """Fit to training data then reduce X to its most important features.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.
            New in v 0.13.0: pandas DataFrames are now also accepted as
            argument for X.
        y : array-like, shape = [n_samples]
            Target values.
            New in v 0.13.0: a pandas Series are now also accepted as
            argument for y.
        groups : array-like, with shape (n_samples,), optional
            Group labels for the samples used while splitting the dataset into
            train/test set. Passed to the fit method of the cross-validator.
        fit_params : various, optional
            Additional parameters that are being passed to the estimator.
            For example, `sample_weights=weights`.

        Returns
        -------
        Reduced feature subset of X, shape={n_samples, k_features}

        """
        self.fit(X, y, groups=groups, **fit_params)
        return self.transform(X)

    def get_metric_dict(self, confidence_interval=0.95):
        """Return metric dictionary

        Parameters
        ----------
        confidence_interval : float (default: 0.95)
            A positive float between 0.0 and 1.0 to compute the confidence
            interval bounds of the CV score averages.

        Returns
        ----------
        Dictionary with items where each dictionary value is a list
        with the number of iterations (number of feature subsets) as
        its length. The dictionary keys corresponding to these lists
        are as follows:
            'feature_idx': tuple of the indices of the feature subset
            'cv_scores': list with individual CV scores
            'avg_score': of CV average scores
            'std_dev': standard deviation of the CV score average
            'std_err': standard error of the CV score average
            'ci_bound': confidence interval bound of the CV score average

        """
        self._check_fitted()
        fdict = deepcopy(self.subsets_)
        for k in fdict:
            std_dev = np.std(self.subsets_[k]["cv_scores"])
            bound, std_err = self._calc_confidence(
                self.subsets_[k]["cv_scores"], confidence=confidence_interval
            )
            fdict[k]["ci_bound"] = bound
            fdict[k]["std_dev"] = std_dev
            fdict[k]["std_err"] = std_err
        return fdict

    def _calc_confidence(self, ary, confidence=0.95):
        std_err = scipy.stats.sem(ary)
        bound = std_err * sp.stats.t._ppf((1 + confidence) / 2.0, len(ary))
        return bound, std_err

    def _check_fitted(self):
        if not self.fitted:
            raise AttributeError(
                "SequentialFeatureSelector has not been" " fitted, yet."
            )



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from .checking import check_Xy, format_kwarg_dictionaries
from .counter import Counter
from .testing import assert_raises

__all__ = ["Counter", "assert_raises", "check_Xy", "format_kwarg_dictionaries"]



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Bootstrap functions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np


def bootstrap(x, func, num_rounds=1000, ci=0.95, ddof=1, seed=None):
    """Implements the ordinary nonparametric bootstrap

    Parameters
    ----------

    x : NumPy array, shape=(n_samples, [n_columns])
        An one or multidimensional array of data records

    func : <func>
        A function which computes a statistic that is used
        to compute the bootstrap replicates (the statistic computed
        from the bootstrap samples). This function must return a
        scalar value. For example, `np.mean` or `np.median` would be
        an acceptable argument for `func` if `x` is a 1-dimensional array
        or vector.

    num_rounds : int (default=1000)
        The number of bootstrap samples to draw where each
        bootstrap sample has the same number of records as the
        original dataset.

    ci : int (default=0.95)
        An integer in the range (0, 1) that represents the
        confidence level for computing the confidence interval.
        For example, `ci=0.95` (default)
        will compute the 95% confidence
        interval from the bootstrap replicates.

    ddof : int
        The delta degrees of freedom used when computing the
        standard error.

    seed : int or None (default=None)
        Random seed for generating bootstrap samples.

    Returns
    -------

    original, standard_error, (lower_ci, upper_ci) : tuple
        Returns the statistic of the original sample (`original`),
        the standard error of the estimate, and the
        respective confidence interval bounds.

    Examples
    --------

    >>> from mlxtend.evaluate import bootstrap
    >>> rng = np.random.RandomState(123)
    >>> x = rng.normal(loc=5., size=100)
    >>> original, std_err, ci_bounds = bootstrap(x,
    ...                                          num_rounds=1000,
    ...                                          func=np.mean,
    ...                                          ci=0.95,
    ...                                          seed=123)
    >>> print('Mean: %.2f, SE: +/- %.2f, CI95: [%.2f, %.2f]' % (original,
    ...                                                         std_err,
    ...                                                         ci_bounds[0],
    ...                                                         ci_bounds[1]))
    Mean: 5.03, SE: +/- 0.11, CI95: [4.80, 5.26]
    >>>

    For more usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/bootstrap/

    """
    if ci <= 0 or ci >= 1:
        raise AttributeError("ci must be in range (0, 1)")

    check_output = func(x)

    if (
        not isinstance(check_output, float)
        and not isinstance(check_output, int)
        and len(check_output.shape) != 0
    ):
        raise AttributeError("func must return a scalar")

    rng = np.random.RandomState(seed)
    bootstrap_replicates = np.zeros(shape=num_rounds)

    # quantile function implemented due
    # to the weird behavior of the NumPy equivalent with
    # either nearest or lower interpolation
    def quantile(x, q):
        rank = round(q * x.shape[0]) - 1
        if rank >= x.shape[0]:
            rank = x.shape[0]
        elif rank <= 0:
            rank = 0
        rank = int(round(rank))
        return x[rank]

    sample_idx = np.arange(x.shape[0])
    for i in range(num_rounds):
        bootstrap_idx = rng.choice(sample_idx, size=sample_idx.shape[0], replace=True)

        bootstrap_replicates[i] = func(x[bootstrap_idx])

    original = check_output
    standard_error = np.std(bootstrap_replicates, ddof=ddof)

    t = np.sort(bootstrap_replicates)
    bound = (1 - ci) / 2.0

    upper_ci = quantile(t, q=(ci + bound))
    lower_ci = quantile(t, q=bound)

    return original, standard_error, (lower_ci, upper_ci)



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
import scipy.stats


def proportion_difference(proportion_1, proportion_2, n_1, n_2=None):
    """
    Computes the test statistic and p-value for a difference of
    proportions test.

    Parameters
    -----------
    proportion_1 : float
        The first proportion
    proportion_2 : float
        The second proportion
    n_1 : int
        The sample size of the first test sample
    n_2 : int or None (default=None)
        The sample size of the second test sample.
        If `None`, `n_1`=`n_2`.

    Returns
    -----------

    z, p : float or None, float
        Returns the z-score and the p-value


    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/proportion_difference/

    """
    if n_2 is None:
        n_2 = n_1

    var_1 = proportion_1 * (1.0 - proportion_1) / n_1
    var_2 = proportion_2 * (1.0 - proportion_2) / n_2

    z = (proportion_1 - proportion_2) / np.sqrt(var_1 + var_2)
    p = scipy.stats.norm.cdf(z)

    return z, p



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# A function for plotting decision regions of classifiers.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import matplotlib.pyplot as plt
import numpy as np


def ecdf(
    x,
    y_label="ECDF",
    x_label=None,
    ax=None,
    percentile=None,
    ecdf_color=None,
    ecdf_marker="o",
    percentile_color="black",
    percentile_linestyle="--",
):
    """Plots an Empirical Cumulative Distribution Function

    Parameters
    ----------
    x : array or list, shape=[n_samples,]
        Array-like object containing the feature values
    y_label : str (default='ECDF')
        Text label for the y-axis
    x_label : str (default=None)
        Text label for the x-axis
    ax : matplotlib.axes.Axes (default: None)
        An existing matplotlib Axes. Creates
        one if ax=None
    percentile : float (default=None)
        Float between 0 and 1 for plotting a percentile
        threshold line
    ecdf_color : matplotlib color (default=None)
        Color for the ECDF plot; uses matplotlib defaults
        if None
    ecdf_marker : matplotlib marker (default='o')
        Marker style for the ECDF plot
    percentile_color : matplotlib color (default='black')
        Color for the percentile threshold if percentile is not None
    percentile_linestyle : matplotlib linestyle (default='--')
        Line style for the percentile threshold if percentile is not None

    Returns
    ---------
    ax : matplotlib.axes.Axes object
    percentile_threshold : float
        Feature threshold at the percentile or None if `percentile=None`
    percentile_count : Number of if percentile is not None
        Number of samples that have a feature less or equal than
        the feature threshold at a percentile threshold
        or None if `percentile=None`

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/plotting/ecdf/

    """
    if ax is None:
        ax = plt.gca()

    x = np.sort(x)
    y = np.arange(1, x.shape[0] + 1) / float(x.shape[0])

    ax.plot(x, y, marker=ecdf_marker, linestyle="", color=ecdf_color)
    ax.set_ylabel("ECDF")
    if x_label is not None:
        ax.set_xlabel(x_label)
    if percentile:
        targets = x[y <= percentile]
        percentile_threshold = targets.max()
        percentile_count = targets.shape[0]
        ax.axvline(
            percentile_threshold, color=percentile_color, linestyle=percentile_linestyle
        )

    else:
        percentile_threshold = None
        percentile_count = None

    return ax, percentile_threshold, percentile_count



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# A function for loading the open-source Iris Flower dataset.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import os

import numpy as np

this_dir, this_filename = os.path.split(__file__)
DATA_PATH = os.path.join(this_dir, "data", "iris.csv.gz")


def iris_data(version="uci"):
    """Iris flower dataset.

    Source : https://archive.ics.uci.edu/ml/datasets/Iris
    Number of samples : 150
    Class labels : {0, 1, 2}, distribution: [50, 50, 50]
        0 = setosa, 1 = versicolor, 2 = virginica.

    Dataset Attributes:

        - 1) sepal length [cm]
        - 2) sepal width [cm]
        - 3) petal length [cm]
        - 4) petal width [cm]


    Parameters
    --------
    version : string, optional (default: 'uci').
      Version to use {'uci', 'corrected'}. 'uci' loads the dataset
      as deposited on the UCI machine learning repository, and
      'corrected' provides the version that is consistent with
      Fisher's original paper. See Note for details.


    Returns
    --------
    X, y : [n_samples, n_features], [n_class_labels]
        X is the feature matrix with 150 flower samples as rows,
        and 4 feature columns sepal length, sepal width,
        petal length, and petal width.
        y is a 1-dimensional array of the class labels {0, 1, 2}


    Note
    --------
    The Iris dataset (originally collected by Edgar Anderson) and
    available in UCI's machine learning repository is different from
    the Iris dataset described in the original paper by  R.A. Fisher [1]).
    Precisely, there are two data points (row number
    34 and 37) in UCI's Machine Learning repository are different from the
    origianlly published Iris dataset. Also, the original version of the Iris
    Dataset, which can be loaded via `version='corrected'` is the same
    as the one in R.

    [1] . A. Fisher (1936). "The use of multiple measurements in taxonomic
    problems". Annals of Eugenics. 7 (2): 179–188

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/data/iris_data/

    """
    if version == "uci":
        tmp = np.genfromtxt(fname=DATA_PATH, delimiter=",")
        X, y = tmp[:, :-1], tmp[:, -1]
        y = y.astype(int)
    elif version == "corrected":
        tmp = np.genfromtxt(fname=DATA_PATH, delimiter=",")
        X, y = tmp[:, :-1], tmp[:, -1]
        X[34] = [4.9, 3.1, 1.5, 0.2]
        X[37] = [4.9, 3.6, 1.4, 0.1]
        y = y.astype(int)
    else:
        raise ValueError("version must be 'uci' or 'corrected'.")
    return X, y



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Linear Discriminant Analysis for dimensionality reduction
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np

from .._base import _BaseModel


class LinearDiscriminantAnalysis(_BaseModel):
    """
    Linear Discriminant Analysis Class

    Parameters
    ----------
    n_discriminants : int (default: None)
        The number of discrimants for transformation.
        Keeps the original dimensions of the dataset if `None`.

    Attributes
    ----------
    w_ : array-like, shape=[n_features, n_discriminants]
        Projection matrix
    e_vals_ : array-like, shape=[n_features]
        Eigenvalues in sorted order.
    e_vecs_ : array-like, shape=[n_features]
       Eigenvectors in sorted order.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/feature_extraction/LinearDiscriminantAnalysis/

    """

    def __init__(self, n_discriminants=None):
        if n_discriminants is not None and n_discriminants < 1:
            raise AttributeError("n_discriminants must be > 1 or None")
        self.n_discriminants = n_discriminants

    def fit(self, X, y, n_classes=None):
        """Fit the LDA model with X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.
        y : array-like, shape = [n_samples]
            Target values.
        n_classes : int (default: None)
            A positive integer to declare the number of class labels
            if not all class labels are present in a partial training set.
            Gets the number of class labels automatically if None.

        Returns
        -------
        self : object

        """
        self._is_fitted = False
        self._check_arrays(X=X, y=y)
        self._fit(X=X, y=y, n_classes=n_classes)
        self._is_fitted = True
        return self

    def _fit(self, X, y, n_classes=None):
        if self.n_discriminants is None or self.n_discriminants > X.shape[1]:
            n_discriminants = X.shape[1]
        else:
            n_discriminants = self.n_discriminants

        if n_classes:
            self._n_classes = n_classes
        else:
            self._n_classes = np.max(y) + 1
        self._n_features = X.shape[1]

        mean_vecs = self._mean_vectors(X=X, y=y, n_classes=self._n_classes)
        within_scatter = self._within_scatter(
            X=X, y=y, n_classes=self._n_classes, mean_vectors=mean_vecs
        )
        between_scatter = self._between_scatter(X=X, y=y, mean_vectors=mean_vecs)
        self.e_vals_, self.e_vecs_ = self._eigendecom(
            within_scatter=within_scatter, between_scatter=between_scatter
        )
        self.w_ = self._projection_matrix(
            eig_vals=self.e_vals_,
            eig_vecs=self.e_vecs_,
            n_discriminants=n_discriminants,
        )
        return self

    def transform(self, X):
        """Apply the linear transformation on X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_projected : np.ndarray, shape = [n_samples, n_discriminants]
            Projected training vectors.

        """
        if not hasattr(self, "w_"):
            raise AttributeError("Object as not been fitted, yet.")
        self._check_arrays(X=X)
        return X.dot(self.w_)

    def _mean_vectors(self, X, y, n_classes):
        mean_vectors = []
        for cl in range(n_classes):
            mean_vectors.append(np.mean(X[y == cl], axis=0))
        return mean_vectors

    def _within_scatter(self, X, y, n_classes, mean_vectors):
        S_W = np.zeros((X.shape[1], X.shape[1]))
        for cl, mv in zip(range(n_classes), mean_vectors):
            class_sc_mat = np.zeros((X.shape[1], X.shape[1]))
            for row in X[y == cl]:
                row, mv = row.reshape(X.shape[1], 1), mv.reshape(X.shape[1], 1)
                class_sc_mat += (row - mv).dot((row - mv).T)
            S_W += class_sc_mat
        return S_W

    def _between_scatter(self, X, y, mean_vectors):
        overall_mean = np.mean(X, axis=0)
        S_B = np.zeros((X.shape[1], X.shape[1]))
        for i, mean_vec in enumerate(mean_vectors):
            n = X[y == i + 1, :].shape[0]
            mean_vec = mean_vec.reshape(X.shape[1], 1)
            overall_mean = overall_mean.reshape(X.shape[1], 1)
            S_B += n * (mean_vec - overall_mean).dot((mean_vec - overall_mean).T)
        return S_B

    def _eigendecom(self, within_scatter, between_scatter):
        e_vals, e_vecs = np.linalg.eig(
            np.linalg.inv(within_scatter).dot(between_scatter)
        )
        sort_idx = np.argsort(e_vals)[::-1]
        e_vals, e_vecs = e_vals[sort_idx], e_vecs[:, sort_idx]
        return e_vals, e_vecs

    def _projection_matrix(self, eig_vals, eig_vecs, n_discriminants):
        matrix_w = np.vstack([eig_vecs[:, i] for i in range(n_discriminants)]).T
        return matrix_w



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from .column_selector import ColumnSelector
from .exhaustive_feature_selector import ExhaustiveFeatureSelector
from .sequential_feature_selector import SequentialFeatureSelector

__all__ = ["ColumnSelector", "SequentialFeatureSelector", "ExhaustiveFeatureSelector"]



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# A function for generating a confusion matrix.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from itertools import product

import numpy as np


def confusion_matrix(y_target, y_predicted, binary=False, positive_label=1):
    """Compute a confusion matrix/contingency table.

    Parameters
    -----------
    y_target : array-like, shape=[n_samples]
        True class labels.
    y_predicted : array-like, shape=[n_samples]
        Predicted class labels.
    binary : bool (default: False)
        Maps a multi-class problem onto a
        binary confusion matrix, where
        the positive class is 1 and
        all other classes are 0.
    positive_label : int (default: 1)
        Class label of the positive class.

    Returns
    ----------
    mat : array-like, shape=[n_classes, n_classes]

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/confusion_matrix/

    """
    if not isinstance(y_target, np.ndarray):
        targ_tmp = np.asarray(y_target)
    else:
        targ_tmp = y_target
    if not isinstance(y_predicted, np.ndarray):
        pred_tmp = np.asarray(y_predicted)
    else:
        pred_tmp = y_predicted

    if len(y_target) != len(y_predicted):
        raise AttributeError(
            "`y_target` and `y_predicted`" " don't have the same number of elements."
        )

    if binary:
        targ_tmp = np.where(targ_tmp != positive_label, 0, 1)
        pred_tmp = np.where(pred_tmp != positive_label, 0, 1)

    class_labels = np.unique(np.concatenate((targ_tmp, pred_tmp)))
    if class_labels.shape[0] == 1:
        if class_labels[0] != 0:
            class_labels = np.array([0, class_labels[0]])
        else:
            class_labels = np.array([class_labels[0], 1])
    n_labels = class_labels.shape[0]
    lst = []
    z = list(zip(targ_tmp, pred_tmp))
    for combi in product(class_labels, repeat=2):
        lst.append(z.count(combi))
    mat = np.asarray(lst)[:, None].reshape(n_labels, n_labels)
    return mat



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Implementation of the logistic regression algorithm for classification.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from time import time

import numpy as np

from .._base import _BaseModel, _Classifier, _IterativeModel


class LogisticRegression(_BaseModel, _IterativeModel, _Classifier):

    """Logistic regression classifier.

    Note that this implementation of Logistic Regression
    expects binary class labels in {0, 1}.

    Parameters
    ------------
    eta : float (default: 0.01)
        Learning rate (between 0.0 and 1.0)

    epochs : int (default: 50)
        Passes over the training dataset.
        Prior to each epoch, the dataset is shuffled
        if `minibatches > 1` to prevent cycles in stochastic gradient descent.

    l2_lambda : float
        Regularization parameter for L2 regularization.
        No regularization if l2_lambda=0.0.

    minibatches : int (default: 1)
        The number of minibatches for gradient-based optimization.
        If 1: Gradient Descent learning
        If len(y): Stochastic Gradient Descent (SGD) online learning
        If 1 < minibatches < len(y): SGD Minibatch learning

    random_seed : int (default: None)
        Set random state for shuffling and initializing the weights.

    print_progress : int (default: 0)
        Prints progress in fitting to stderr.
        0: No output
        1: Epochs elapsed and cost
        2: 1 plus time elapsed
        3: 2 plus estimated time until completion

    Attributes
    -----------
    w_ : 2d-array, shape={n_features, 1}
      Model weights after fitting.

    b_ : 1d-array, shape={1,}
      Bias unit after fitting.

    cost_ : list
        List of floats with cross_entropy cost (sgd or gd) for every
        epoch.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/classifier/LogisticRegression/

    """

    def __init__(
        self,
        eta=0.01,
        epochs=50,
        l2_lambda=0.0,
        minibatches=1,
        random_seed=None,
        print_progress=0,
    ):
        _BaseModel.__init__(self)
        _IterativeModel.__init__(self)
        _Classifier.__init__(self)

        self.eta = eta
        self.epochs = epochs
        self.l2_lambda = l2_lambda
        self.minibatches = minibatches
        self.random_seed = random_seed
        self.print_progress = print_progress
        self._is_fitted = False

    def _forward(self, X):
        z = self._net_input(X)
        a = self._sigmoid_activation(z)
        return a

    def _backward(self, X, y_true, y_probas):
        grad_loss_wrt_out = y_true - y_probas
        grad_loss_wrt_w = -X.T @ grad_loss_wrt_out.reshape(-1, 1)
        grad_loss_wrt_b = -np.sum(grad_loss_wrt_out)
        return grad_loss_wrt_w, grad_loss_wrt_b

    def _fit(self, X, y, init_params=True):
        self._check_target_array(y, allowed={(0, 1)})

        if init_params:
            self.b_, self.w_ = self._init_params(
                weights_shape=(X.shape[1], 1),
                bias_shape=(1,),
                random_seed=self.random_seed,
            )
            self.cost_ = []

        self.init_time_ = time()
        rgen = np.random.RandomState(self.random_seed)
        for i in range(self.epochs):
            for idx in self._yield_minibatches_idx(
                rgen=rgen, n_batches=self.minibatches, data_ary=y, shuffle=True
            ):
                y_val = self._forward(X[idx])
                grad_loss_wrt_w, grad_loss_wrt_b = self._backward(
                    X[idx], y_true=y[idx], y_probas=y_val
                )

                l2_reg = self.l2_lambda * self.w_
                self.w_ += self.eta * (-grad_loss_wrt_w - l2_reg)
                self.b_ += self.eta * -grad_loss_wrt_b.sum()

            cost = self._logit_cost(y, self._forward(X))
            self.cost_.append(cost)
            if self.print_progress:
                self._print_progress(iteration=(i + 1), n_iter=self.epochs, cost=cost)
        return self

    def _predict(self, X):
        # equivalent to np.where(self._forward(X) < 0.5, 0, 1)
        return np.where(self._net_input(X) < 0.0, 0, 1)

    def _net_input(self, X):
        """Compute the linear net input."""
        return (X.dot(self.w_) + self.b_).flatten()

    def predict_proba(self, X):
        """Predict class probabilities of X from the net input.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        ----------
        Class 1 probability : float

        """
        return self._forward(X)

    def _logit_cost(self, y, y_val):
        logit = -y.dot(np.log(y_val)) - ((1 - y).dot(np.log(1 - y_val)))
        if self.l2_lambda:
            l2 = self.l2_lambda / 2.0 * np.sum(self.w_**2)
            logit += l2
        return logit

    def _sigmoid_activation(self, z):
        """Compute the output of the logistic sigmoid function."""
        return 1.0 / (1.0 + np.exp(-z))



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# A function for searching files in local directories.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import os


def find_files(
    substring,
    path,
    recursive=False,
    check_ext=None,
    ignore_invisible=True,
    ignore_substring=None,
):
    """Find files in a directory based on substring matching.

    Parameters
    ----------
    substring : `str`
        Substring of the file to be matched.
    path : `str`
        Path where to look.
    recursive : `bool`
        If true, searches subdirectories recursively.
    check_ext : `str`
        If string (e.g., '.txt'), only returns files that
        match the specified file extension.
    ignore_invisible : `bool`
        If `True`, ignores invisible files
        (i.e., files starting with a period).
    ignore_substring : `str`
        Ignores files that contain the specified substring.

    Returns
    ----------
    results : `list`
        List of the matched files.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/file_io/find_files/

    """

    def check_file(f, path):
        if not (ignore_substring and ignore_substring in f):
            if substring in f:
                compl_path = os.path.join(path, f)
                if os.path.isfile(compl_path):
                    return compl_path
        return False

    results = []

    if recursive:
        for par, nxt, fnames in os.walk(path):
            for f in fnames:
                fn = check_file(f, par)
                if fn:
                    results.append(fn)

    else:
        for f in os.listdir(path):
            if ignore_invisible and f.startswith("."):
                continue
            fn = check_file(f, path)
            if fn:
                results.append(fn)

    if check_ext:
        results = [r for r in results if os.path.splitext(r)[-1] == check_ext]

    return results



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Matplotlib wrapper for generating stacked barplots.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause


from itertools import cycle

import matplotlib.pyplot as plt
import numpy as np


def stacked_barplot(
    df,
    bar_width="auto",
    colors="bgrcky",
    labels="index",
    rotation=90,
    legend_loc="best",
):
    """
    Function to plot stacked barplots

    Parameters
    ----------
    df : pandas.DataFrame
        A pandas DataFrame where the index denotes the
        x-axis labels, and the columns contain the different
        measurements for each row.
    bar_width: 'auto' or float (default: 'auto')
        Parameter to set the widths of the bars. if
        'auto', the width is automatically determined by
        the number of columns in the dataset.
    colors: str (default: 'bgrcky')
        The colors of the bars.
    labels: 'index' or iterable (default: 'index')
        If 'index', the DataFrame index will be used as
        x-tick labels.
    rotation: int (default: 90)
        Parameter to rotate the x-axis labels.
    legend_loc : str (default: 'best')
        Location of the plot legend
        {best, upper left, upper right, lower left, lower right}
        No legend if legend_loc=False

    Returns
    ---------
    fig : matplotlib.pyplot figure object

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/plotting/stacked_barplot/

    """
    # Setting the positions and width for the bars
    pos = np.array(range(len(df.index)))

    if bar_width == "auto":
        width = 1 / (len(df.columns[1:]) * 2)
    else:
        width = bar_width

    if labels == "index":
        labels = df.index

    color_gen = cycle(colors)

    label_pos = [pos]

    # Plotting the bars
    fig, ax = plt.subplots(figsize=(12, 6))

    plt.bar(
        pos, df.iloc[:, 0], width, alpha=0.8, color=next(color_gen), label=df.columns[0]
    )

    for i, c in enumerate(df.columns[1:]):
        bar_pos = [p + width * (i + 1) for p in pos]
        label_pos.append(bar_pos)
        plt.bar(
            bar_pos, df.iloc[:, i + 1], width, alpha=0.5, color=next(color_gen), label=c
        )

    # Setting axis labels and ticks

    label_pos = np.asarray(label_pos).mean(axis=0) + width * 0.5

    ax.set_xticks(label_pos)
    ax.set_xticklabels(labels, rotation=rotation, horizontalalignment="center")

    # Setting the x-axis and y-axis limits
    plt.xlim(min(pos) - width, max(pos) + width * 7)

    if legend_loc:
        plt.legend(loc=legend_loc, scatterpoints=1)
    return fig



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Functions for different counting operations.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause


def factorial(n):
    if n == 0:
        return 1
    else:
        return n * factorial(n - 1)


def num_combinations(n, k, with_replacement=False):
    """
    Function to calculate the number of possible combinations.

    Parameters
    ----------
    n : `int`
        Total number of items.
    k : `int`
        Number of elements of the target itemset.
    with_replacement : `bool` (default: False)
        Allows repeated elements if True.

    Returns
    ----------
    comb : `int`
        Number of possible combinations.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/math/num_combinations/

    """
    if with_replacement:
        numerator = factorial(n + k - 1)
        denominator = factorial(k) * factorial(n - 1)
    else:
        numerator = factorial(n)
        denominator = factorial(k) * factorial(n - k)
    comb = numerator // denominator
    return comb


def num_permutations(n, k, with_replacement=False):
    """
    Function to calculate the number of possible permutations.

    Parameters
    ----------
    n : `int`
      Total number of items.
    k : `int`
      Number of elements of the target itemset.
    with_replacement : `bool`
      Allows repeated elements if True.

    Returns
    ----------
    permut : `int`
      Number of possible permutations.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/math/num_permutations/

    """
    if with_replacement:
        permut = n**k
    else:
        numerator = factorial(n)
        denominator = factorial(n - k)
        permut = numerator // denominator
    return permut



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Base Clusteer (Clutering Parent Class)
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np


class _MultiLayer(object):
    def __init__(self):
        pass

    def _layermapping(self, n_features, n_classes, hidden_layers):
        """Creates a dictionaries of layer dimensions for weights and biases.

        For example, given
        `n_features=10`, `n_classes=10`, and `hidden_layers=[8, 7, 6]`:

        biases =
           {1: [[8], 'n_hidden_1'],
            2: [[7], 'n_hidden_2'],
            3: [[6], 'n_hidden_3'],
           'out': [[10], 'n_classes']
           }

        weights =
           {1: [[10, 8], 'n_features, n_hidden_1'],
            2: [[8, 7], 'n_hidden_1, n_hidden_2'],
            3: [[7, 6], 'n_hidden_2, n_hidden_3'],
            'out': [[6, 10], 'n_hidden_3, n_classes']
            }

        """
        weights = {
            "1": [[n_features, hidden_layers[0]], "n_features, n_hidden_1"],
            "out": [
                [hidden_layers[-1], n_classes],
                "n_hidden_%d, n_classes" % len(hidden_layers),
            ],
        }
        biases = {
            "1": [[hidden_layers[0]], "n_hidden_1"],
            "out": [[n_classes], "n_classes"],
        }

        if len(hidden_layers) > 1:
            for i, h in enumerate(hidden_layers[1:]):
                layer = i + 2
                weights[str(layer)] = [
                    [weights[str(layer - 1)][0][1], h],
                    "n_hidden_%d, n_hidden_%d" % (layer - 1, layer),
                ]
                biases[str(layer)] = [[h], "n_hidden_%d" % layer]
        return weights, biases

    def _init_params_from_layermapping(self, weight_maps, bias_maps, random_seed=None):
        rgen = np.random.RandomState(random_seed)
        weights, biases = {}, {}

        rgen = np.random.RandomState(random_seed)
        weights, biases = {}, {}

        for kw, kb in zip(sorted(weight_maps), sorted(bias_maps)):
            weights[kw] = rgen.normal(loc=0.0, scale=0.01, size=weight_maps[kw][0])
            biases[kb] = np.zeros(shape=bias_maps[kb][0])
        return weights, biases



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np


def one_hot(y, num_labels="auto", dtype="float"):
    """One-hot encoding of class labels

    Parameters
    ----------
    y : array-like, shape = [n_classlabels]
        Python list or numpy array consisting of class labels.
    num_labels : int or 'auto'
        Number of unique labels in the class label array. Infers the number
        of unique labels from the input array if set to 'auto'.
    dtype : str
        NumPy array type (float, float32, float64) of the output array.

    Returns
    ----------
    ary : numpy.ndarray, shape = [n_classlabels]
        One-hot encoded array, where each sample is represented as
        a row vector in the returned array.

    Examples
    ----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/preprocessing/one_hot/

    """
    if not (num_labels == "auto" or isinstance(num_labels, int)):
        raise AttributeError('num_labels must be an integer or "auto"')
    if isinstance(y, list):
        yt = np.asarray(y)
    else:
        yt = y
    if not len(yt.shape) == 1:
        raise AttributeError("y array must be 1-dimensional")
    if num_labels == "auto":
        # uniq = np.unique(yt).shape[0]
        uniq = np.max(yt + 1)
    else:
        uniq = num_labels
    if uniq == 1:
        ary = np.array([[0.0]], dtype=dtype)

    else:
        ary = np.zeros((len(y), uniq))
        for i, val in enumerate(y):
            ary[i, val] = 1

    return ary.astype(dtype)



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Implementation of a Multi-layer Perceptron in Tensorflow
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from time import time

import numpy as np
from scipy.special import expit

from .._base import _BaseModel, _Classifier, _IterativeModel, _MultiClass, _MultiLayer


class MultiLayerPerceptron(
    _BaseModel, _IterativeModel, _MultiClass, _MultiLayer, _Classifier
):

    """Multi-layer perceptron classifier with logistic sigmoid activations

    Parameters
    ------------
    eta : float (default: 0.5)
        Learning rate (between 0.0 and 1.0)
    epochs : int (default: 50)
        Passes over the training dataset.
        Prior to each epoch, the dataset is shuffled
        if `minibatches > 1` to prevent cycles in stochastic gradient descent.
    hidden_layers : list (default: [50])
        Number of units per hidden layer. By default 50 units in the
        first hidden layer. At the moment only 1 hidden layer is supported
    n_classes : int (default: None)
        A positive integer to declare the number of class labels
        if not all class labels are present in a partial training set.
        Gets the number of class labels automatically if None.
    l1 : float (default: 0.0)
        L1 regularization strength
    l2 : float (default: 0.0)
        L2 regularization strength
    momentum : float (default: 0.0)
        Momentum constant. Factor multiplied with the
        gradient of the previous epoch t-1 to improve
        learning speed
        w(t) := w(t) - (grad(t) + momentum * grad(t-1))
    decrease_const : float (default: 0.0)
        Decrease constant. Shrinks the learning rate
        after each epoch via eta / (1 + epoch*decrease_const)
    minibatches : int (default: 1)
        Divide the training data into *k* minibatches
        for accelerated stochastic gradient descent learning.
        Gradient Descent Learning if `minibatches` = 1
        Stochastic Gradient Descent learning if `minibatches` = len(y)
        Minibatch learning if `minibatches` > 1
    random_seed : int (default: None)
        Set random state for shuffling and initializing the weights.
    print_progress : int (default: 0)
        Prints progress in fitting to stderr.
        0: No output
        1: Epochs elapsed and cost
        2: 1 plus time elapsed
        3: 2 plus estimated time until completion

    Attributes
    -----------
    w_ : 2d-array, shape=[n_features, n_classes]
        Weights after fitting.
    b_ : 1D-array, shape=[n_classes]
        Bias units after fitting.
    cost_ : list
        List of floats; the mean categorical cross entropy
        cost after each epoch.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/classifier/MultiLayerPerceptron/

    """

    def __init__(
        self,
        eta=0.5,
        epochs=50,
        hidden_layers=[50],
        n_classes=None,
        momentum=0.0,
        l1=0.0,
        l2=0.0,
        dropout=1.0,
        decrease_const=0.0,
        minibatches=1,
        random_seed=None,
        print_progress=0,
    ):
        _BaseModel.__init__(self)
        _Classifier.__init__(self)
        _IterativeModel.__init__(self)
        _MultiClass.__init__(self)
        _MultiLayer.__init__(self)

        if len(hidden_layers) > 1:
            raise AttributeError("Currently, only 1 hidden layer is supported")
        self.hidden_layers = hidden_layers
        self.eta = eta
        self.n_classes = n_classes
        self.l1 = l1
        self.l2 = l2
        self.decrease_const = decrease_const
        self.momentum = momentum
        self.epochs = epochs
        self.minibatches = minibatches
        self.random_seed = random_seed
        self.print_progress = print_progress
        self._is_fitted = False

    def _fit(self, X, y, init_params=True):
        self._check_target_array(y)

        if init_params:
            self._decr_eta = self.eta
            if self.n_classes is None:
                self.n_classes = np.max(y) + 1

            self._n_features = X.shape[1]
            self._weight_maps, self._bias_maps = self._layermapping(
                n_features=self._n_features,
                n_classes=self.n_classes,
                hidden_layers=self.hidden_layers,
            )

            self.w_, self.b_ = self._init_params_from_layermapping(
                weight_maps=self._weight_maps,
                bias_maps=self._bias_maps,
                random_seed=self.random_seed,
            )

            self.cost_ = []

            if self.momentum != 0.0:
                prev_grad_b_1 = np.zeros(shape=self.b_["1"].shape)
                prev_grad_w_1 = np.zeros(shape=self.w_["1"].shape)
                prev_grad_b_out = np.zeros(shape=self.b_["out"].shape)
                prev_grad_w_out = np.zeros(shape=self.w_["out"].shape)

        y_enc = self._one_hot(y=y, n_labels=self.n_classes, dtype=np.float_)

        self.init_time_ = time()

        rgen = np.random.RandomState(self.random_seed)
        for i in range(self.epochs):
            for idx in self._yield_minibatches_idx(
                rgen=rgen, n_batches=self.minibatches, data_ary=y, shuffle=True
            ):
                net_1, act_1, net_out, act_out = self._feedforward(X[idx])

                # GRADIENTS VIA BACKPROPAGATION

                # [n_samples, n_classlabels]
                sigma_out = act_out - y_enc[idx]

                # [n_samples, n_hidden]
                sigmoid_derivative_1 = act_1 * (1.0 - act_1)

                # [n_samples, n_classlabels] dot [n_classlabels, n_hidden]
                # -> [n_samples, n_hidden]
                sigma_1 = np.dot(sigma_out, self.w_["out"].T) * sigmoid_derivative_1

                # [n_features, n_samples] dot [n_samples, n_hidden]
                # -> [n_features, n_hidden]
                grad_W_1 = np.dot(X[idx].T, sigma_1)

                grad_B_1 = np.sum(sigma_1, axis=0)

                # [n_hidden, n_samples] dot [n_samples, n_classlabels]
                # -> [n_hidden, n_classlabels]
                grad_W_out = np.dot(act_1.T, sigma_out)

                grad_B_out = np.sum(sigma_out, axis=0)

                # LEARNING RATE ADJUSTEMENTS
                self._decr_eta /= 1.0 + self.decrease_const * i

                # REGULARIZATION AND WEIGHT UPDATES

                dW_1 = (
                    self._decr_eta * grad_W_1 + self._decr_eta * self.l2 * self.w_["1"]
                )

                dW_out = (
                    self._decr_eta * grad_W_out
                    + self._decr_eta * self.l2 * self.w_["out"]
                )

                dB_1 = self._decr_eta * grad_B_1
                dB_out = self._decr_eta * grad_B_out

                self.w_["1"] -= dW_1
                self.b_["1"] -= dB_1
                self.w_["out"] -= dW_out
                self.b_["out"] -= dB_out

                if self.momentum != 0.0:
                    self.w_["1"] -= self.momentum * prev_grad_w_1
                    self.b_["1"] -= self.momentum * prev_grad_b_1
                    self.w_["out"] -= self.momentum * prev_grad_w_out
                    self.b_["out"] -= self.momentum * prev_grad_b_out
                    prev_grad_b_1 = grad_B_1
                    prev_grad_w_1 = grad_W_1
                    prev_grad_b_out = grad_B_out
                    prev_grad_w_out = grad_W_out

            net_1, act_1, net_out, act_out = self._feedforward(X)
            cross_ent = self._cross_entropy(output=act_out, y_target=y_enc)
            cost = self._compute_cost(cross_ent)

            self.cost_.append(cost)
            if self.print_progress:
                self._print_progress(iteration=i + 1, n_iter=self.epochs, cost=cost)

        return self

    def _feedforward(self, X):
        # [n_samples, n_features] dot [n_features, n_hidden]
        # -> [n_samples, n_hidden]
        net_1 = np.dot(X, self.w_["1"]) + self.b_["1"]
        act_1 = self._sigmoid(net_1)

        # [n_samples, n_hidden] dot [n_hidden, n_classlabels]
        # -> [n_samples, n_classlabels]
        net_out = np.dot(act_1, self.w_["out"]) + self.b_["out"]
        act_out = self._softmax(net_out)

        return net_1, act_1, net_out, act_out

    def _compute_cost(self, cross_entropy):
        L2_term = self.l2 * (
            np.sum(self.w_["1"] ** 2.0) + np.sum(self.w_["out"] ** 2.0)
        )

        L1_term = self.l1 * (np.abs(self.w_["1"]).sum() + np.abs(self.w_["out"]).sum())

        cross_entropy = cross_entropy + L2_term + L1_term
        return 0.5 * np.mean(cross_entropy)

    def _predict(self, X):
        net_1, act_1, net_out, act_out = self._feedforward(X)
        y_pred = np.argmax(net_out, axis=1)
        return y_pred

    def _softmax(self, z):
        e_x = np.exp(z - z.max(axis=1, keepdims=True))
        out = e_x / e_x.sum(axis=1, keepdims=True)
        return out
        # return (np.exp(z.T) / np.sum(np.exp(z), axis=1)).T

    def _cross_entropy(self, output, y_target):
        return -np.sum(np.log(output) * (y_target), axis=1)

    def predict_proba(self, X):
        """Predict class probabilities of X from the net input.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        ----------
        Class probabilties : array-like, shape= [n_samples, n_classes]

        """
        net_1, act_1, net_out, act_out = self._feedforward(X)
        softm = self._softmax(act_out)
        return softm

    def _sigmoid(self, z):
        """Compute logistic function (sigmoid).
        Uses scipy.special.expit to avoid overflow
        error for very small input values z.
        """
        # return 1.0 / (1.0 + np.exp(-z))
        return expit(z)



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# A function for plotting enrichment plots.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause


from itertools import cycle

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def enrichment_plot(
    df,
    colors="bgrkcy",
    markers=" ",
    linestyles="-",
    alpha=0.5,
    lw=2,
    where="post",
    grid=True,
    count_label="Count",
    xlim="auto",
    ylim="auto",
    invert_axes=False,
    legend_loc="best",
    ax=None,
):
    """Plot stacked barplots

    Parameters
    ----------
    df : pandas.DataFrame
        A pandas DataFrame where columns represent the different categories.
    colors: str (default: 'bgrcky')
        The colors of the bars.
    markers : str (default: ' ')
        Matplotlib markerstyles, e.g,
        'sov' for square,circle, and triangle markers.
    linestyles : str (default: '-')
        Matplotlib linestyles, e.g.,
        '-,--' to cycle normal and dashed lines. Note
        that the different linestyles need to be separated by commas.
    alpha : float (default: 0.5)
        Transparency level from 0.0 to 1.0.
    lw : int or float (default: 2)
        Linewidth parameter.
    where : {'post', 'pre', 'mid'} (default: 'post')
        Starting location of the steps.
    grid : bool (default: `True`)
        Plots a grid if True.
    count_label : str (default: 'Count')
        Label for the "Count"-axis.
    xlim : 'auto' or array-like [min, max] (default: 'auto')
        Min and maximum position of the x-axis range.
    ylim : 'auto' or array-like [min, max] (default: 'auto')
        Min and maximum position of the y-axis range.
    invert_axes : bool (default: False)
        Plots count on the x-axis if True.
    legend_loc : str (default: 'best')
        Location of the plot legend
        {best, upper left, upper right, lower left, lower right}
        No legend if legend_loc=False
    ax : matplotlib axis, optional (default: None)
        Use this axis for plotting or make a new one otherwise

    Returns
    ----------
    ax : matplotlib axis

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/plotting/enrichment_plot/

    """
    if isinstance(df, pd.Series):
        df_temp = pd.DataFrame(df)
    else:
        df_temp = df

    if ax is None:
        ax = plt.gca()

    color_gen = cycle(colors)
    marker_gen = cycle(markers)
    linestyle_gen = cycle(linestyles.split(","))
    r = range(1, len(df_temp.index) + 1)
    labels = df_temp.columns

    x_data = df_temp
    y_data = r

    for lab in labels:
        x, y = sorted(x_data[lab]), y_data
        if invert_axes:
            x, y = y, x

        ax.step(
            x,
            y,
            where=where,
            label=lab,
            color=next(color_gen),
            alpha=alpha,
            lw=lw,
            marker=next(marker_gen),
            linestyle=next(linestyle_gen),
        )

    if invert_axes:
        ax.set_ylim, ax.set_xlim = ax.set_xlim, ax.set_ylim

    if ylim == "auto":
        ax.set_ylim([np.min(y_data) - 1, np.max(y_data) + 1])
    else:
        ax.set_ylim(ylim)

    if xlim == "auto":
        df_min, df_max = np.min(x_data.min()), np.max(x_data.max())
        ax.set_xlim([df_min - 1, df_max + 1])

    else:
        ax.set_xlim(xlim)

    if legend_loc:
        plt.legend(loc=legend_loc, scatterpoints=1)

    if grid:
        plt.grid()

    if count_label:
        if invert_axes:
            plt.xlabel(count_label)
        else:
            plt.ylabel(count_label)

    return ax



# Sebastian Raschka 2014-2023
# contributor: Vahid Mirjalili
# mlxtend Machine Learning Library Extensions
#
# A class for transforming face images.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import os
import warnings

import numpy as np
from skimage.transform import AffineTransform, resize, warp

from ..externals.pyprind.progbar import ProgBar
from . import extract_face_landmarks
from .utils import read_image

LEFT_INDEX = np.array([36, 37, 38, 39, 40, 41])
RIGHT_INDEX = np.array([42, 43, 44, 45, 46, 47])


class EyepadAlign(object):
    """Class to align/transform face images to facial landmarks,
       based on eye alignment.

       1. A scaling factor is computed based on distance between the
        left and right eye, such that the transformed face image will
        have the same eye distance as a reference face image.

       2. A transformation is performed based on the eyes' center point.
        to align the face based on the reference eye location.

       3. Finally, the transformed image is padded with zeros to match
        the desired final image size.

    Parameters
    ----------
    verbose : int (default=0)
        Verbose level to display the progress bar and log messages.
        Setting `verbose=1` will print a progress bar upon calling
        `fit_directory`.

    Attributes
    ----------
    target_landmarks_ : target landmarks to transform new face images to.
        Depending on the chosen `fit` parameters, it can  be either
          (1) assigned to pre-fit shapes,
          (2) computed from a single face image
          (3) computed as the mean of face landmarks
              from all face images in a file directory of face images.

    eye_distance_ : the distance between left and right eyes
        in the target landmarks.

    target_height_ : the height of the transformed output image.

    target_width_ : the width of the transformed output image.

    Returns
    -------
    self : object

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/image/eyepad_align/

    """

    def __init__(self, verbose=0):
        self.verbose = verbose

    def fit_image(self, target_image):
        """Derives facial landmarks from a target image.

        Arguments
        ----------
        target_image : `uint8` numpy.array, shape=[width, height, channels]
            NumPy array representation of the image data.

        Returns
        -------
        self : object

        """
        landmarks = extract_face_landmarks(target_image)
        self.target_landmarks_ = landmarks
        self.target_width_ = target_image.shape[1]
        self.target_height_ = target_image.shape[0]

        props = self._calc_eye_properties(self.target_landmarks_)
        self.eyes_mid_point_ = props[0]
        self.eye_distance_ = props[1]
        return self

    def fit_directory(
        self,
        target_img_dir,
        target_height,
        target_width,
        file_extension=".jpg",
        pre_check=True,
    ):
        """
        Calculates the average landmarks for all face images
        in a directory which will then be set as the target landmark set.

        Arguments
        ----------
        target_img_dir : str
            Directory containing the images

         target_height : int
            Expected image height of the images in the directory

        target_width : int
            Expected image width of the images in the directory

        file_extension str (default='.jpg'): File extension of the image files.

        pre_check Bool (default=True): Checks that each image has the
            dimensions specificed via `target_height`
            and `target_width` on the whole directory first to identify
            potential issues that are recommended
            to be fixed before proceeding. Raises a warning for each image if
            dimensions differ from the ones specified and expected.

        Returns
        -------
        self : object

        """
        self.target_height_ = target_height
        self.target_width_ = target_width

        file_list = [
            os.path.relpath(os.path.join(dirpath, f), target_img_dir)
            for (dirpath, dirnames, filenames) in os.walk(target_img_dir)
            for f in filenames
            if f.endswith(file_extension)
        ]

        if not len(file_list):
            raise ValueError(
                "No images found in %s with extension %s."
                % (target_img_dir, file_extension)
            )

        landmarks_list = []

        if pre_check:
            if self.verbose >= 1:
                print("Pre-Checking directory for" " consistent image dimensions...")
                pbar = ProgBar(len(file_list))
            for f in file_list:
                img = read_image(filename=f, path=target_img_dir)
                if self.verbose >= 1:
                    pbar.update()
                if (
                    img.shape[0] != self.target_height_
                    or img.shape[1] != self.target_width_
                ):
                    warnings.warn(
                        "Image %s has "
                        "dimensions %d x %d "
                        "instead of %d x %d."
                        % (
                            f,
                            img.shape[0],
                            img.shape[1],
                            self.target_height_,
                            self.target_width_,
                        )
                    )

        if self.verbose >= 1:
            print(
                "Fitting the average facial landmarks "
                "for %d face images " % (len(file_list))
            )
            pbar = ProgBar(len(file_list))
        for f in file_list:
            img = read_image(filename=f, path=target_img_dir)
            if self.verbose >= 1:
                pbar.update()

            if self.target_width_ != img.shape[1]:
                width_ratio = self.target_width_ / img.shape[1]
                height_ratio = self.target_height_ / img.shape[0]

                if np.abs(width_ratio - height_ratio) > 0.001:  # ignore
                    continue

                img = resize(
                    img,
                    output_shape=(self.target_height_, self.target_width_),
                    anti_aliasing=True,
                    mode="reflect",
                )
                img = (img * 255).astype("uint8")

            landmarks = extract_face_landmarks(img)
            if np.sum(landmarks) is not None:  # i.e., None == no face detected
                landmarks_list.append(landmarks)
            else:
                warnings.warn("No face detected in image %s. Image ignored." % f)
        self.target_landmarks_ = np.mean(landmarks_list, axis=0)

        props = self._calc_eye_properties(self.target_landmarks_)
        self.eyes_mid_point_ = props[0]
        self.eye_distance_ = props[1]
        return self

    def fit_values(self, target_landmarks, target_width, target_height):
        """Used for determining the eye location from pre-defined
            landmark arrays, eliminating the need for re-computing
               the average landmarks on a target image or image directory.

        Arguments
        ---------
        target_landmarks : np.array, shape=(height, width)
            NumPy array containing the locations of the facial landmarks
            as determined by `mlxtend.image.extract_face_landmarks`

        target_height : int
            image height

        target_width : int
            image width

        Returns
        -------
        self : object

        """
        self.target_landmarks_ = target_landmarks
        self.target_width_ = target_width
        self.target_height_ = target_height

        props = self._calc_eye_properties(self.target_landmarks_)
        self.eyes_mid_point_ = props[0]
        self.eye_distance_ = props[1]
        return self

    def _calc_eye_properties(self, landmarks):
        """Calculates the face properties:
        (1) coordinates of the left-eye
        (2) coordinates of the right-eye
        (3) the distance between left and right eyes
        (4) the middle point between the two eyes

        """
        left_eye = np.mean(landmarks[LEFT_INDEX], axis=0)
        right_eye = np.mean(landmarks[RIGHT_INDEX], axis=0)
        eyes_mid_point = (left_eye + right_eye) / 2.0
        eye_distance = np.sqrt(np.sum(np.square(left_eye - right_eye)))

        return eyes_mid_point, eye_distance

    def transform(self, img):
        """transforms a single face image (img) to the target landmarks
               based on the location of the eyes by
               scaling, translation and cropping (if needed):

            (1) Scaling the image so that the distance of the two eyes
                in the given image (img) matches the distance of the
                two eyes in the target landmarks.

            (2) Translation is performed based on the middle point
                between the two eyes.

        Arguments
        ---------

        img : np.array, shape=(height, width, channels)
            Input image to be transformed.

        Returns
        -------
        self : object

        """

        if not hasattr(self, "eyes_mid_point_"):
            raise AttributeError(
                "Not fitted, yet. Call one of the `fit*`"
                " methods prior to using `transform`."
            )

        landmarks = extract_face_landmarks(img)
        if landmarks is None:
            return

        eyes_mid_point, eye_distance = self._calc_eye_properties(landmarks)

        scale = self.eye_distance_ / eye_distance
        tr = self.eyes_mid_point_ / scale - eyes_mid_point
        tr = (int(tr[0] * scale), int(tr[1] * scale))

        tform = AffineTransform(
            scale=(scale, scale), rotation=0, shear=0, translation=tr
        )
        h, w = self.target_height_, self.target_width_
        img_tr = warp(img, tform.inverse, output_shape=(h, w))
        return np.array(img_tr * 255, dtype="uint8")



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Function for plotting linear regression fits via scikit-learn and matplotlib.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause


import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import pearsonr
from sklearn.linear_model import LinearRegression


def plot_linear_regression(
    X,
    y,
    model=LinearRegression(),
    corr_func="pearsonr",
    scattercolor="blue",
    fit_style="k--",
    legend=True,
    xlim="auto",
):
    """Plot a linear regression line fit.

    Parameters
    ----------
    X : numpy array, shape = [n_samples,]
        Samples.
    y : numpy array, shape (n_samples,)
        Target values
    model: object (default: sklearn.linear_model.LinearRegression)
        Estimator object for regression. Must implement
        a .fit() and .predict() method.
    corr_func: str or function (default: 'pearsonr')
        Uses `pearsonr` from scipy.stats if corr_func='pearsonr'.
        to compute the regression slope. If not 'pearsonr', the `corr_func`,
        the `corr_func` parameter expects a function of the form
        func(<x-array>, <y-array>) as inputs, which is expected to return
        a tuple `(<correlation_coefficient>, <some_unused_value>)`.
    scattercolor: string (default: blue)
        Color of scatter plot points.
    fit_style: string (default: k--)
        Style for the line fit.
    legend: bool (default: True)
        Plots legend with corr_coeff coef.,
        fit coef., and intercept values.
    xlim: array-like (x_min, x_max) or 'auto' (default: 'auto')
        X-axis limits for the linear line fit.

    Returns
    ----------
    regression_fit : tuple
        intercept, slope, corr_coeff (float, float, float)

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/plotting/plot_linear_regression/

    """
    if isinstance(X, list):
        X = np.asarray(X, dtype=float)
    if isinstance(y, list):
        y = np.asarray(y, dtype=float)
    if len(X.shape) == 1:
        X = X[:, np.newaxis]

    model.fit(X, y)

    plt.scatter(X, y, c=scattercolor)

    if xlim == "auto":
        x_min, x_max = X[:, 0].min(), X[:, 0].max()
        x_min -= 0.2 * x_min
        x_max += 0.2 * x_max

    else:
        x_min, x_max = xlim

    y_min = model.predict(x_min)
    y_max = model.predict(x_max)

    plt.plot([x_min, x_max], [y_min, y_max], fit_style, lw=1)

    if corr_func == "pearsonr":
        corr_func = pearsonr

    corr_coeff, p = corr_func(X[:, 0], y)
    intercept, slope = model.intercept_, model.coef_[0]

    if legend:
        leg_text = "intercept: %.2f\nslope: %.2f" % (intercept, slope)
        if corr_func:
            leg_text += "\ncorrelation: %.2f" % corr_coeff
        plt.legend([leg_text], loc="best")
    regression_fit = (intercept, slope, corr_coeff)
    return regression_fit



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
from scipy import stats
from sklearn.metrics import get_scorer
from sklearn.model_selection import KFold, train_test_split


def paired_ttest_resampled(
    estimator1,
    estimator2,
    X,
    y,
    num_rounds=30,
    test_size=0.3,
    scoring=None,
    random_seed=None,
):
    """
    Implements the resampled paired t test procedure
    to compare the performance of two models
    (also called k-hold-out paired t test).

    Parameters
    ----------
    estimator1 : scikit-learn classifier or regressor

    estimator2 : scikit-learn classifier or regressor

    X : {array-like, sparse matrix}, shape = [n_samples, n_features]
        Training vectors, where n_samples is the number of samples and
        n_features is the number of features.

    y : array-like, shape = [n_samples]
        Target values.

    num_rounds : int (default: 30)
        Number of resampling iterations
        (i.e., train/test splits)

    test_size : float or int (default: 0.3)
        If float, should be between 0.0 and 1.0 and
        represent the proportion of the dataset to use
        as a test set.
        If int, represents the absolute number of test exsamples.

    scoring : str, callable, or None (default: None)
        If None (default), uses 'accuracy' for sklearn classifiers
        and 'r2' for sklearn regressors.
        If str, uses a sklearn scoring metric string identifier, for example
        {accuracy, f1, precision, recall, roc_auc} for classifiers,
        {'mean_absolute_error', 'mean_squared_error'/'neg_mean_squared_error',
        'median_absolute_error', 'r2'} for regressors.
        If a callable object or function is provided, it has to be conform with
        sklearn's signature ``scorer(estimator, X, y)``; see
        https://scikit-learn.org/stable/modules/generated/sklearn.metrics.make_scorer.html
        for more information.

    random_seed : int or None (default: None)
        Random seed for creating the test/train splits.

    Returns
    ----------
    t : float
        The t-statistic

    pvalue : float
        Two-tailed p-value.
        If the chosen significance level is larger
        than the p-value, we reject the null hypothesis
        and accept that there are significant differences
        in the two compared models.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/paired_ttest_resampled/

    """
    if not isinstance(test_size, int) and not isinstance(test_size, float):
        raise ValueError(
            "train_size must be of " "type int or float. Got %s." % type(test_size)
        )

    rng = np.random.RandomState(random_seed)

    if scoring is None:
        if estimator1._estimator_type == "classifier":
            scoring = "accuracy"
        elif estimator1._estimator_type == "regressor":
            scoring = "r2"
        else:
            raise AttributeError("Estimator must " "be a Classifier or Regressor.")
    if isinstance(scoring, str):
        scorer = get_scorer(scoring)
    else:
        scorer = scoring

    score_diff = []
    for i in range(num_rounds):
        randint = rng.randint(low=0, high=32767)

        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=test_size, random_state=randint
        )

        estimator1.fit(X_train, y_train)
        estimator2.fit(X_train, y_train)

        est1_score = scorer(estimator1, X_test, y_test)
        est2_score = scorer(estimator2, X_test, y_test)
        score_diff.append(est1_score - est2_score)

    avg_diff = np.mean(score_diff)

    numerator = avg_diff * np.sqrt(num_rounds)
    denominator = np.sqrt(
        sum([(diff - avg_diff) ** 2 for diff in score_diff]) / (num_rounds - 1)
    )
    t_stat = numerator / denominator

    pvalue = stats.t.sf(np.abs(t_stat), num_rounds - 1) * 2.0
    return float(t_stat), float(pvalue)


def paired_ttest_kfold_cv(
    estimator1, estimator2, X, y, cv=10, scoring=None, shuffle=False, random_seed=None
):
    """
    Implements the k-fold paired t test procedure
    to compare the performance of two models.

    Parameters
    ----------
    estimator1 : scikit-learn classifier or regressor

    estimator2 : scikit-learn classifier or regressor

    X : {array-like, sparse matrix}, shape = [n_samples, n_features]
        Training vectors, where n_samples is the number of samples and
        n_features is the number of features.

    y : array-like, shape = [n_samples]
        Target values.

    cv : int (default: 10)
        Number of splits and iteration for the
        cross-validation procedure

    scoring : str, callable, or None (default: None)
        If None (default), uses 'accuracy' for sklearn classifiers
        and 'r2' for sklearn regressors.
        If str, uses a sklearn scoring metric string identifier, for example
        {accuracy, f1, precision, recall, roc_auc} for classifiers,
        {'mean_absolute_error', 'mean_squared_error'/'neg_mean_squared_error',
        'median_absolute_error', 'r2'} for regressors.
        If a callable object or function is provided, it has to be conform with
        sklearn's signature ``scorer(estimator, X, y)``; see
        https://scikit-learn.org/stable/modules/generated/sklearn.metrics.make_scorer.html
        for more information.

    shuffle : bool (default: True)
        Whether to shuffle the dataset for generating
        the k-fold splits.

    random_seed : int or None (default: None)
        Random seed for shuffling the dataset
        for generating the k-fold splits.
        Ignored if shuffle=False.

    Returns
    ----------
    t : float
        The t-statistic

    pvalue : float
        Two-tailed p-value.
        If the chosen significance level is larger
        than the p-value, we reject the null hypothesis
        and accept that there are significant differences
        in the two compared models.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/paired_ttest_kfold_cv/

    """

    if not shuffle:
        kf = KFold(n_splits=cv, shuffle=shuffle)
    else:
        kf = KFold(n_splits=cv, random_state=random_seed, shuffle=shuffle)

    if scoring is None:
        if estimator1._estimator_type == "classifier":
            scoring = "accuracy"
        elif estimator1._estimator_type == "regressor":
            scoring = "r2"
        else:
            raise AttributeError("Estimator must " "be a Classifier or Regressor.")
    if isinstance(scoring, str):
        scorer = get_scorer(scoring)
    else:
        scorer = scoring

    score_diff = []

    for train_index, test_index in kf.split(X):
        X_train, X_test = X[train_index], X[test_index]
        y_train, y_test = y[train_index], y[test_index]

        estimator1.fit(X_train, y_train)
        estimator2.fit(X_train, y_train)

        est1_score = scorer(estimator1, X_test, y_test)
        est2_score = scorer(estimator2, X_test, y_test)
        score_diff.append(est1_score - est2_score)

    avg_diff = np.mean(score_diff)

    numerator = avg_diff * np.sqrt(cv)
    denominator = np.sqrt(
        sum([(diff - avg_diff) ** 2 for diff in score_diff]) / (cv - 1)
    )
    t_stat = numerator / denominator

    pvalue = stats.t.sf(np.abs(t_stat), cv - 1) * 2.0
    return float(t_stat), float(pvalue)


def paired_ttest_5x2cv(estimator1, estimator2, X, y, scoring=None, random_seed=None):
    """
    Implements the 5x2cv paired t test proposed
    by Dieterrich (1998)
    to compare the performance of two models.

    Parameters
    ----------
    estimator1 : scikit-learn classifier or regressor

    estimator2 : scikit-learn classifier or regressor

    X : {array-like, sparse matrix}, shape = [n_samples, n_features]
        Training vectors, where n_samples is the number of samples and
        n_features is the number of features.

    y : array-like, shape = [n_samples]
        Target values.

    scoring : str, callable, or None (default: None)
        If None (default), uses 'accuracy' for sklearn classifiers
        and 'r2' for sklearn regressors.
        If str, uses a sklearn scoring metric string identifier, for example
        {accuracy, f1, precision, recall, roc_auc} for classifiers,
        {'mean_absolute_error', 'mean_squared_error'/'neg_mean_squared_error',
        'median_absolute_error', 'r2'} for regressors.
        If a callable object or function is provided, it has to be conform with
        sklearn's signature ``scorer(estimator, X, y)``; see
        https://scikit-learn.org/stable/modules/generated/sklearn.metrics.make_scorer.html
        for more information.

    random_seed : int or None (default: None)
        Random seed for creating the test/train splits.

    Returns
    ----------
    t : float
        The t-statistic

    pvalue : float
        Two-tailed p-value.
        If the chosen significance level is larger
        than the p-value, we reject the null hypothesis
        and accept that there are significant differences
        in the two compared models.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/paired_ttest_5x2cv/

    """
    rng = np.random.RandomState(random_seed)

    if scoring is None:
        if estimator1._estimator_type == "classifier":
            scoring = "accuracy"
        elif estimator1._estimator_type == "regressor":
            scoring = "r2"
        else:
            raise AttributeError("Estimator must " "be a Classifier or Regressor.")
    if isinstance(scoring, str):
        scorer = get_scorer(scoring)
    else:
        scorer = scoring

    variance_sum = 0.0
    first_diff = None

    def score_diff(X_1, X_2, y_1, y_2):
        estimator1.fit(X_1, y_1)
        estimator2.fit(X_1, y_1)
        est1_score = scorer(estimator1, X_2, y_2)
        est2_score = scorer(estimator2, X_2, y_2)
        score_diff = est1_score - est2_score
        return score_diff

    for i in range(5):
        randint = rng.randint(low=0, high=32767)
        X_1, X_2, y_1, y_2 = train_test_split(X, y, test_size=0.5, random_state=randint)

        score_diff_1 = score_diff(X_1, X_2, y_1, y_2)
        score_diff_2 = score_diff(X_2, X_1, y_2, y_1)
        score_mean = (score_diff_1 + score_diff_2) / 2.0
        score_var = (score_diff_1 - score_mean) ** 2 + (score_diff_2 - score_mean) ** 2
        variance_sum += score_var
        if first_diff is None:
            first_diff = score_diff_1

    numerator = first_diff
    denominator = np.sqrt(1 / 5.0 * variance_sum)
    t_stat = numerator / denominator

    pvalue = stats.t.sf(np.abs(t_stat), 5) * 2.0
    return float(t_stat), float(pvalue)



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause


from .accuracy import accuracy_score
from .bias_variance_decomp import bias_variance_decomp
from .bootstrap import bootstrap
from .bootstrap_outofbag import BootstrapOutOfBag
from .bootstrap_point632 import bootstrap_point632_score
from .cochrans_q import cochrans_q
from .confusion_matrix import confusion_matrix
from .counterfactual import create_counterfactual
from .f_test import combined_ftest_5x2cv, ftest
from .feature_importance import feature_importance_permutation
from .holdout import PredefinedHoldoutSplit, RandomHoldoutSplit
from .lift_score import lift_score
from .mcnemar import mcnemar, mcnemar_table, mcnemar_tables
from .permutation import permutation_test
from .proportion_difference import proportion_difference
from .scoring import scoring
from .time_series import GroupTimeSeriesSplit
from .ttest import paired_ttest_5x2cv, paired_ttest_kfold_cv, paired_ttest_resampled

__all__ = [
    "scoring",
    "confusion_matrix",
    "mcnemar_table",
    "mcnemar_tables",
    "mcnemar",
    "lift_score",
    "bootstrap",
    "permutation_test",
    "BootstrapOutOfBag",
    "bootstrap_point632_score",
    "cochrans_q",
    "paired_ttest_resampled",
    "paired_ttest_kfold_cv",
    "paired_ttest_5x2cv",
    "feature_importance_permutation",
    "RandomHoldoutSplit",
    "PredefinedHoldoutSplit",
    "ftest",
    "combined_ftest_5x2cv",
    "proportion_difference",
    "bias_variance_decomp",
    "accuracy_score",
    "create_counterfactual",
    "GroupTimeSeriesSplit",
]



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Bootstrap functions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np


class BootstrapOutOfBag(object):
    """
    Parameters
    ----------

    n_splits : int (default=200)
        Number of bootstrap iterations.
        Must be larger than 1.

    random_seed : int (default=None)
        If int, random_seed is the seed used by
        the random number generator.


    Returns
    -------
    train_idx : ndarray
        The training set indices for that split.

    test_idx : ndarray
        The testing set indices for that split.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/BootstrapOutOfBag/

    """

    def __init__(self, n_splits=200, random_seed=None):
        self.random_seed = random_seed

        if not isinstance(n_splits, int) or n_splits < 1:
            raise ValueError("Number of splits must be greater than 1.")
        self.n_splits = n_splits

    def split(self, X, y=None, groups=None):
        """

        y : array-like or None (default: None)
            Argument is not used and only included as parameter
            for compatibility, similar to `KFold` in scikit-learn.

        groups : array-like or None (default: None)
            Argument is not used and only included as parameter
            for compatibility, similar to `KFold` in scikit-learn.


        """
        rng = np.random.RandomState(self.random_seed)
        sample_idx = np.arange(X.shape[0])
        set_idx = set(sample_idx)

        for _ in range(self.n_splits):
            train_idx = rng.choice(sample_idx, size=sample_idx.shape[0], replace=True)
            test_idx = np.array(list(set_idx - set(train_idx)))
            yield train_idx, test_idx

    def get_n_splits(self, X=None, y=None, groups=None):
        """Returns the number of splitting iterations in the cross-validator

        Parameters
        ----------
        X : object
            Always ignored, exists for compatibility with scikit-learn.

        y : object
            Always ignored, exists for compatibility with scikit-learn.

        groups : object
            Always ignored, exists for compatibility with scikit-learn.

        Returns
        -------

        n_splits : int
            Returns the number of splitting iterations in the cross-validator.

        """
        return self.n_splits



# Sebastian Raschka 2014-2023
# contributor: Vahid Mirjalili
# mlxtend Machine Learning Library Extensions
#
# A counter class for printing the progress of an iterator.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import bz2
import os
import tarfile
import zipfile

import imageio


def check_exists(path):
    path = os.path.expanduser(path)
    return os.path.exists(path)


def makedir(path):
    path = os.path.expanduser(path)
    if not check_exists(path):
        os.makedirs(path)


def listdir(path, extensions=""):
    path = os.path.expanduser(path)
    if check_exists(path):
        return [f for f in os.listdir(path) if f.endswith(extensions)]
    else:
        raise FileNotFoundError


def read_image(filename, path=None):
    if path is not None:
        path = os.path.expanduser(path)
        filename = os.path.join(path, filename)
    if check_exists(filename):
        return imageio.imread(filename)
    else:
        raise FileNotFoundError


def download_url(url, save_path):
    from six.moves import urllib

    save_path = os.path.expanduser(save_path)
    if not check_exists(save_path):
        makedir(save_path)

    filename = url.rpartition("/")[2]
    filepath = os.path.join(save_path, filename)

    try:
        print("Downloading " + url + " to " + filepath)
        urllib.request.urlretrieve(url, filepath)
    except ValueError:
        raise Exception(
            "Failed to download! Check URL: " + url + " and local path: " + save_path
        )


def extract_file(path, to_directory=None):
    path = os.path.expanduser(path)
    if path.endswith(".zip"):
        opener, mode = zipfile.ZipFile, "r"
    elif path.endswith((".tar.gz", ".tgz")):
        opener, mode = tarfile.open, "r:gz"
    elif path.endswith(("tar.bz2", ".tbz")):
        opener, mode = tarfile.open, "r:bz2"
    elif path.endswith(".bz2"):
        opener, mode = bz2.BZ2File, "rb"
        with open(path[:-4], "wb") as fp_out, opener(path, "rb") as fp_in:
            for data in iter(lambda: fp_in.read(100 * 1024), b""):
                fp_out.write(data)
        return
    else:
        raise (
            ValueError,
            "Could not extract `{}` as no extractor is found!".format(path),
        )

    if to_directory is None:
        to_directory = os.path.abspath(os.path.join(path, os.path.pardir))
    cwd = os.getcwd()
    os.chdir(to_directory)

    try:
        file = opener(path, mode)
        try:
            file.extractall()
        finally:
            file.close()
    finally:
        os.chdir(cwd)



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Classes for column-based scaling of datasets
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
import pandas as pd


def minmax_scaling(array, columns, min_val=0, max_val=1):
    """Min max scaling of pandas' DataFrames.

    Parameters
    ----------
    array : pandas DataFrame or NumPy ndarray, shape = [n_rows, n_columns].
    columns : array-like, shape = [n_columns]
        Array-like with column names, e.g., ['col1', 'col2', ...]
        or column indices [0, 2, 4, ...]
    min_val : `int` or `float`, optional (default=`0`)
        minimum value after rescaling.
    max_val : `int` or `float`, optional (default=`1`)
        maximum value after rescaling.

    Returns
    ----------
    df_new : pandas DataFrame object.
        Copy of the array or DataFrame with rescaled columns.

    Examples
    ----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/preprocessing/minmax_scaling/

    """
    ary_new = array.astype(float)
    if len(ary_new.shape) == 1:
        ary_new = ary_new[:, np.newaxis]

    if isinstance(ary_new, pd.DataFrame):
        ary_newt = ary_new.loc
    elif isinstance(ary_new, np.ndarray):
        ary_newt = ary_new
    else:
        raise AttributeError("Input array must be a pandas" "DataFrame or NumPy array")

    numerator = ary_newt[:, columns] - ary_newt[:, columns].min(axis=0)
    denominator = ary_newt[:, columns].max(axis=0) - ary_newt[:, columns].min(axis=0)
    ary_newt[:, columns] = numerator / denominator

    if not min_val == 0 and not max_val == 1:
        ary_newt[:, columns] = ary_newt[:, columns] * (max_val - min_val) + min_val

    return ary_newt[:, columns]


def standardize(array, columns=None, ddof=0, return_params=False, params=None):
    """Standardize columns in pandas DataFrames.

    Parameters
    ----------
    array : pandas DataFrame or NumPy ndarray, shape = [n_rows, n_columns].
    columns : array-like, shape = [n_columns] (default: None)
        Array-like with column names, e.g., ['col1', 'col2', ...]
        or column indices [0, 2, 4, ...]
        If None, standardizes all columns.
    ddof : int (default: 0)
        Delta Degrees of Freedom. The divisor used in calculations
        is N - ddof, where N represents the number of elements.
    return_params : dict (default: False)
        If set to True, a dictionary is returned in addition to the
        standardized array. The parameter dictionary contains the
        column means ('avgs') and standard deviations ('stds') of
        the individual columns.
    params : dict (default: None)
        A dictionary with column means and standard deviations as
        returned by the `standardize` function if `return_params`
        was set to True. If a `params` dictionary is provided, the
        `standardize` function will use these instead of computing
        them from the current array.

    Notes
    ----------
    If all values in a given column are the same, these values are all
    set to `0.0`. The standard deviation in the `parameters` dictionary
    is consequently set to `1.0` to avoid dividing by zero.

    Returns
    ----------
    df_new : pandas DataFrame object.
        Copy of the array or DataFrame with standardized columns.

    Examples
    ----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/preprocessing/standardize/

    """
    ary_new = array.astype(float)
    dim = ary_new.shape
    if len(dim) == 1:
        ary_new = ary_new[:, np.newaxis]

    if isinstance(ary_new, pd.DataFrame):
        ary_newt = ary_new.loc
        if columns is None:
            columns = ary_new.columns
    elif isinstance(ary_new, np.ndarray):
        ary_newt = ary_new
        if columns is None:
            columns = list(range(ary_new.shape[1]))

    else:
        raise AttributeError("Input array must be a pandas " "DataFrame or NumPy array")

    if params is not None:
        parameters = params
    else:
        parameters = {
            "avgs": ary_newt[:, columns].mean(axis=0),
            "stds": ary_newt[:, columns].std(axis=0, ddof=ddof),
        }
    are_constant = np.all(ary_newt[:, columns] == ary_newt[0, columns], axis=0)

    for c, b in zip(columns, are_constant):
        if b:
            ary_newt[:, c] = np.zeros(dim[0])
            parameters["stds"][c] = 1.0

    ary_newt[:, columns] = (ary_newt[:, columns] - parameters["avgs"]) / parameters[
        "stds"
    ]

    if return_params:
        return ary_newt[:, columns], parameters
    else:
        return ary_newt[:, columns]



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from itertools import combinations

import numpy as np
import scipy.stats


def mcnemar_table(y_target, y_model1, y_model2):
    """
    Compute a 2x2 contigency table for McNemar's test.

    Parameters
    -----------
    y_target : array-like, shape=[n_samples]
        True class labels as 1D NumPy array.
    y_model1 : array-like, shape=[n_samples]
        Predicted class labels from model as 1D NumPy array.
    y_model2 : array-like, shape=[n_samples]
        Predicted class labels from model 2 as 1D NumPy array.

    Returns
    ----------
    tb : array-like, shape=[2, 2]
       2x2 contingency table with the following contents:
       a: tb[0, 0]: # of samples that both models predicted correctly
       b: tb[0, 1]: # of samples that model 1 got right and model 2 got wrong
       c: tb[1, 0]: # of samples that model 2 got right and model 1 got wrong
       d: tb[1, 1]: # of samples that both models predicted incorrectly

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/mcnemar_table/

    """
    for ary in (y_target, y_model1, y_model2):
        if len(ary.shape) != 1:
            raise ValueError("One or more input arrays are not 1-dimensional.")

    if y_target.shape[0] != y_model1.shape[0]:
        raise ValueError(
            "y_target and y_model1 contain a different number" " of elements."
        )

    if y_target.shape[0] != y_model2.shape[0]:
        raise ValueError(
            "y_target and y_model2 contain a different number" " of elements."
        )

    m1_vs_true = (y_target == y_model1).astype(int)
    m2_vs_true = (y_target == y_model2).astype(int)

    plus_true = m1_vs_true + m2_vs_true
    minus_true = m1_vs_true - m2_vs_true

    tb = np.zeros((2, 2), dtype=int)

    tb[0, 0] = np.sum(plus_true == 2)
    tb[0, 1] = np.sum(minus_true == 1)
    tb[1, 0] = np.sum(minus_true == -1)
    tb[1, 1] = np.sum(plus_true == 0)

    return tb


def mcnemar_tables(y_target, *y_model_predictions):
    """
    Compute multiple 2x2 contigency tables for McNemar's
    test or Cochran's Q test.

    Parameters
    -----------
    y_target : array-like, shape=[n_samples]
        True class labels as 1D NumPy array.

    y_model_predictions : array-like, shape=[n_samples]
        Predicted class labels for a model.

    Returns
    ----------

    tables : dict
        Dictionary of NumPy arrays with shape=[2, 2]. Each dictionary
        key names the two models to be compared based on the order the
        models were passed as `*y_model_predictions`. The number of
        dictionary entries is equal to the number of pairwise combinations
        between the m models, i.e., "m choose 2."

        For example the following target array (containing the true labels)
        and 3 models

        - y_true = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
        - y_mod0 = np.array([0, 1, 0, 0, 0, 1, 1, 0, 0, 0])
        - y_mod1 = np.array([0, 0, 1, 1, 0, 1, 1, 0, 0, 0])
        - y_mod2 = np.array([0, 1, 1, 1, 0, 1, 0, 0, 0, 0])

        would result in the following dictionary:


        {'model_0 vs model_1': array([[ 4.,  1.],
                                      [ 2.,  3.]]),
         'model_0 vs model_2': array([[ 3.,  0.],
                                      [ 3.,  4.]]),
         'model_1 vs model_2': array([[ 3.,  0.],
                                      [ 2.,  5.]])}

        Each array is structured in the following way:

        - tb[0, 0]: # of samples that both models predicted correctly
        - tb[0, 1]: # of samples that model a got right and model b got wrong
        - tb[1, 0]: # of samples that model b got right and model a got wrong
        - tb[1, 1]: # of samples that both models predicted incorrectly

    Examples
    -----------

    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/mcnemar_tables/

    """
    model_lens = set()
    y_model_predictions = list(y_model_predictions)
    for ary in [y_target] + y_model_predictions:
        if len(ary.shape) != 1:
            raise ValueError("One or more input arrays are not 1-dimensional.")
        model_lens.add(ary.shape[0])

    if len(model_lens) > 1:
        raise ValueError(
            "Each prediction array must have the " "same number of samples."
        )

    num_models = len(y_model_predictions)

    if num_models < 2:
        raise ValueError("Provide at least 2 model prediction arrays.")

    tables = {}

    for comb in combinations(range(num_models), 2):
        tb = np.zeros((2, 2))
        model1_vs_true = (y_target == y_model_predictions[comb[0]]).astype(int)
        model2_vs_true = (y_target == y_model_predictions[comb[1]]).astype(int)
        plus_true = model1_vs_true + model2_vs_true
        minus_true = model1_vs_true - model2_vs_true

        tb[0, 0] = np.sum(plus_true == 2)
        tb[0, 1] = np.sum(minus_true == 1)
        tb[1, 0] = np.sum(minus_true == -1)
        tb[1, 1] = np.sum(plus_true == 0)

        name_str = "model_%s vs model_%s" % (comb[0], comb[1])
        tables[name_str] = tb

    return tables


def mcnemar(ary, corrected=True, exact=False):
    """
    McNemar test for paired nominal data

    Parameters
    -----------
    ary : array-like, shape=[2, 2]
        2 x 2 contigency table (as returned by evaluate.mcnemar_table),
        where
        a: ary[0, 0]: # of samples that both models predicted correctly
        b: ary[0, 1]: # of samples that model 1 got right and model 2 got wrong
        c: ary[1, 0]: # of samples that model 2 got right and model 1 got wrong
        d: aryCell [1, 1]: # of samples that both models predicted incorrectly
    corrected : array-like, shape=[n_samples] (default: True)
        Uses Edward's continuity correction for chi-squared if `True`
    exact : bool, (default: False)
        If `True`, uses an exact binomial test comparing b to
        a binomial distribution with n = b + c and p = 0.5.
        It is highly recommended to use `exact=True` for sample sizes < 25
        since chi-squared is not well-approximated
        by the chi-squared distribution!

    Returns
    -----------
    chi2, p : float or None, float
        Returns the chi-squared value and the p-value;
        if `exact=True` (default: `False`), `chi2` is `None`

    Examples
    -----------

    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/mcnemar/

    """

    if not ary.shape == (2, 2):
        raise ValueError("Input array must be a 2x2 array.")

    b = ary[0, 1]
    c = ary[1, 0]
    n = b + c

    if not exact:
        if corrected:
            chi2 = (abs(ary[0, 1] - ary[1, 0]) - 1.0) ** 2 / float(n)
        else:
            chi2 = (ary[0, 1] - ary[1, 0]) ** 2 / float(n)
        p = scipy.stats.distributions.chi2.sf(chi2, 1)

    else:
        chi2 = None
        p = min(scipy.stats.binom.cdf(min(b, c), b + c, 0.5) * 2.0, 1.0)

        # this is equivalent to the following code:
        #
        #    p = 0
        #    for i in range(max(b, c), b+c+1):
        #        p += scipy.special.binom(b+c, i) * 0.5**i * (1-0.5)**((b+c)-i)
        #    p = 2*p

    return chi2, p



"""Utilties to handle estimator list"""

from sklearn.utils.metaestimators import _BaseComposition


class _BaseXComposition(_BaseComposition):
    """
    parameter handler for list of estimators
    """

    def _set_params(self, attr, named_attr, **params):
        # Ordered parameter replacement
        # 1. root parameter
        if attr in params:
            setattr(self, attr, params.pop(attr))

        # 2. single estimator replacement
        items = getattr(self, named_attr)
        names = []
        if items:
            names, estimators = zip(*items)
            estimators = list(estimators)
        for name in list(params.keys()):
            if "__" not in name and name in names:
                # replace single estimator and re-build the
                # root estimators list
                for i, est_name in enumerate(names):
                    if est_name == name:
                        new_val = params.pop(name)
                        if new_val is None:
                            del estimators[i]
                        else:
                            estimators[i] = new_val
                        break
                # replace the root estimators
                setattr(self, attr, estimators)

        # 3. estimator parameters and other initialisation arguments
        super(_BaseXComposition, self).set_params(**params)
        return self



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from .linear_discriminant_analysis import LinearDiscriminantAnalysis
from .principal_component_analysis import PrincipalComponentAnalysis
from .rbf_kernel_pca import RBFKernelPCA

__all__ = ["PrincipalComponentAnalysis", "LinearDiscriminantAnalysis", "RBFKernelPCA"]



"""
Sebastian Raschka 2014-2017
Python Progress Indicator Utility

Author: Sebastian Raschka <sebastianraschka.com>
License: BSD 3 clause

Contributors: https://github.com/rasbt/pyprind/graphs/contributors
Code Repository: https://github.com/rasbt/pyprind
PyPI: https://pypi.python.org/pypi/PyPrind
"""


import os
import sys
import time
from io import UnsupportedOperation

try:
    import psutil

    psutil_import = True
except ImportError:
    psutil_import = False


class Prog:
    def __init__(
        self, iterations, track_time, stream, title, monitor, update_interval=None
    ):
        """Initializes tracking object."""
        self.cnt = 0
        self.title = title
        self.max_iter = float(iterations)  # to support Python 2.x
        self.track = track_time
        self.start = time.time()
        self.end = None
        self.item_id = None
        self.eta = None
        self.total_time = 0.0
        self.last_time = self.start
        self.monitor = monitor
        self.stream = stream
        self.active = True
        self._stream_out = self._no_stream
        self._stream_flush = self._no_stream
        self._check_stream()
        self._print_title()
        self.update_interval = update_interval
        self._cached_output = ""

        sys.stdout.flush()
        sys.stderr.flush()

        if monitor:
            if not psutil_import:
                raise ValueError(
                    "psutil package is required when using" " the `monitor` option."
                )
            else:
                self.process = psutil.Process()
        if self.track:
            self.eta = 1

    def update(self, iterations=1, item_id=None, force_flush=False):
        """
        Updates the progress bar / percentage indicator.

        Parameters
        ----------
        iterations : int (default: 1)
            default argument can be changed to integer values
            >=1 in order to update the progress indicators more than once
            per iteration.
        item_id : str (default: None)
            Print an item_id sring behind the progress bar
        force_flush : bool (default: False)
            If True, flushes the progress indicator to the output screen
            in each iteration.

        """
        self.item_id = item_id
        self.cnt += iterations
        self._print(force_flush=force_flush)
        self._finish()

    def stop(self):
        """Stops the progress bar / percentage indicator if necessary."""
        self.cnt = self.max_iter
        self._finish()

    def _check_stream(self):
        """Determines which output stream (stdout, stderr, or custom) to use"""
        if self.stream:
            try:
                supported = "PYCHARM_HOSTED" in os.environ or os.isatty(
                    sys.stdout.fileno()
                )

            # a fix for IPython notebook "IOStream has no fileno."
            except UnsupportedOperation:
                supported = True

            else:
                if self.stream is not None and hasattr(self.stream, "write"):
                    self._stream_out = self.stream.write
                    self._stream_flush = self.stream.flush

            if supported:
                if self.stream == 1:
                    self._stream_out = sys.stdout.write
                    self._stream_flush = sys.stdout.flush
                elif self.stream == 2:
                    self._stream_out = sys.stderr.write
                    self._stream_flush = sys.stderr.flush
            else:
                if self.stream is not None and hasattr(self.stream, "write"):
                    self._stream_out = self.stream.write
                    self._stream_flush = self.stream.flush
                else:
                    print("Warning: No valid output stream.")

    def _elapsed(self):
        """Returns elapsed time at update."""
        self.last_time = time.time()
        return self.last_time - self.start

    def _calc_eta(self):
        """Calculates estimated time left until completion."""
        elapsed = self._elapsed()
        if self.cnt == 0 or elapsed < 0.001:
            return None
        rate = float(self.cnt) / elapsed
        self.eta = (float(self.max_iter) - float(self.cnt)) / rate

    def _calc_percent(self):
        """Calculates the rel. progress in percent with 2 decimal points."""
        return round(self.cnt / self.max_iter * 100, 2)

    def _no_stream(self, text=None):
        """Called when no valid output stream is available."""
        pass

    def _get_time(self, _time):
        if _time < 86400:
            return time.strftime("%H:%M:%S", time.gmtime(_time))
        else:
            s = (
                str(int(_time // 3600))
                + ":"
                + time.strftime("%M:%S", time.gmtime(_time))
            )
            return s

    def _finish(self):
        """Determines if maximum number of iterations (seed) is reached."""
        if self.active and self.cnt >= self.max_iter:
            self.total_time = self._elapsed()
            self.end = time.time()
            self.last_progress -= 1  # to force a refreshed _print()
            self._print()
            if self.track:
                self._stream_out(
                    "\nTotal time elapsed: " + self._get_time(self.total_time)
                )
            self._stream_out("\n")
            self.active = False

    def _print_title(self):
        """Prints tracking title at initialization."""
        if self.title:
            self._stream_out("{}\n".format(self.title))
            self._stream_flush()

    def _cache_eta(self):
        """Prints the estimated time left."""
        self._calc_eta()
        self._cached_output += " | ETA: " + self._get_time(self.eta)

    def _cache_item_id(self):
        """Prints an item id behind the tracking object."""
        self._cached_output += " | Item ID: %s" % self.item_id

    def __repr__(self):
        str_start = time.strftime("%m/%d/%Y %H:%M:%S", time.localtime(self.start))
        str_end = time.strftime("%m/%d/%Y %H:%M:%S", time.localtime(self.end))
        self._stream_flush()

        time_info = (
            "Title: {}\n"
            "  Started: {}\n"
            "  Finished: {}\n"
            "  Total time elapsed: ".format(self.title, str_start, str_end)
            + self._get_time(self.total_time)
        )
        if self.monitor:
            try:
                cpu_total = self.process.cpu_percent()
                mem_total = self.process.memory_percent()
            except AttributeError:  # old version of psutil
                cpu_total = self.process.get_cpu_percent()
                mem_total = self.process.get_memory_percent()

            cpu_mem_info = "  CPU %: {:.2f}\n" "  Memory %: {:.2f}".format(
                cpu_total, mem_total
            )

            return time_info + "\n" + cpu_mem_info
        else:
            return time_info

    def __str__(self):
        return self.__repr__()



# Sebastian Raschka 2014-2023
# mlxtend Machine Learning Library Extensions
#
# Base Regressor (Regressor Parent Class)
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from time import time

import numpy as np

from .._base import _BaseModel, _IterativeModel, _Regressor


class LinearRegression(_BaseModel, _IterativeModel, _Regressor):

    """Ordinary least squares linear regression.

    Parameters
    ------------
    method : string (default: 'direct')
        For gradient descent-based optimization, use `sgd` (see `minibatch`
        parameter for further options). Otherwise, if `direct` (default),
        the analytical method is used. For alternative, numerically more
        stable solutions, use either `qr` (QR decomopisition) or `svd`
        (Singular Value Decomposition).
    eta : float (default: 0.01)
        solver learning rate (between 0.0 and 1.0). Used with `method =`
        `'sgd'`. (See `methods` parameter for details)
    epochs : int (default: 50)
        Passes over the training dataset.
        Prior to each epoch, the dataset is shuffled
        if `minibatches > 1` to prevent cycles in stochastic gradient descent.
        Used with `method = 'sgd'`. (See `methods` parameter for details)
    minibatches : int (default: None)
        The number of minibatches for gradient-based optimization.
        If None: Direct method, QR, or SVD method (see `method` parameter
                 for details)
        If 1: Gradient Descent learning
        If len(y): Stochastic Gradient Descent learning
        If 1 < minibatches < len(y): Minibatch learning
    random_seed : int (default: None)
        Set random state for shuffling and initializing the weights. Used in
        `method = 'sgd'`. (See `methods` parameter for details)
    print_progress : int (default: 0)
        Prints progress in fitting to stderr if `method = 'sgd'`.
        0: No output
        1: Epochs elapsed and cost
        2: 1 plus time elapsed
        3: 2 plus estimated time until completion

    Attributes
    -----------
    w_ : 2d-array, shape={n_features, 1}
      Model weights after fitting.
    b_ : 1d-array, shape={1,}
      Bias unit after fitting.
    cost_ : list
        Sum of squared errors after each epoch;
        ignored if solver='normal equation'

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/regressor/LinearRegression/

    """

    def __init__(
        self,
        method="direct",
        eta=0.01,
        epochs=50,
        minibatches=None,
        random_seed=None,
        print_progress=0,
    ):
        _BaseModel.__init__(self)
        _IterativeModel.__init__(self)
        _Regressor.__init__(self)
        self.eta = eta
        self.epochs = epochs
        self.minibatches = minibatches
        self.random_seed = random_seed
        self.print_progress = print_progress
        self._is_fitted = False
        self.method = method

        if method != "sgd" and minibatches is not None:
            raise ValueError(
                (
                    "Minibatches should be set to `None` "
                    "if `method` != `sgd`. Got method=`%s`."
                )
                % (method)
            )

        supported_methods = ("sgd", "direct", "svd", "qr")
        if method not in supported_methods:
            raise ValueError(
                "`method` must be in %s. Got %s." % (supported_methods, method)
            )

    def _fit(self, X, y, init_params=True):
        if init_params:
            self.b_, self.w_ = self._init_params(
                weights_shape=(X.shape[1], 1),
                bias_shape=(1,),
                random_seed=self.random_seed,
            )
            self.cost_ = []

        # Direct analytical method
        if self.method == "direct":
            self.b_, self.w_ = self._normal_equation(X, y)
        # Gradient descent or stochastic gradient descent learning
        elif self.method == "sgd":
            self.init_time_ = time()
            rgen = np.random.RandomState(self.random_seed)
            for i in range(self.epochs):
                for idx in self._yield_minibatches_idx(
                    rgen=rgen, n_batches=self.minibatches, data_ary=y, shuffle=True
                ):
                    y_val = self._net_input(X[idx])
                    errors = y[idx] - y_val
                    self.w_ += self.eta * X[idx].T.dot(errors).reshape(self.w_.shape)
                    self.b_ += self.eta * errors.sum()

                cost = self._sum_squared_error_cost(y, self._net_input(X))
                self.cost_.append(cost)
                if self.print_progress:
                    self._print_progress(
                        iteration=(i + 1), n_iter=self.epochs, cost=cost
                    )
        # Solve using QR decomposition
        elif self.method == "qr":
            Xb = np.hstack((np.ones((X.shape[0], 1)), X))
            Q, R = np.linalg.qr(Xb)
            beta = np.dot(np.linalg.inv(R), np.dot(Q.T, y))
            self.b_ = np.array([beta[0]])
            self.w_ = beta[1:].reshape(X.shape[1], 1)
        # Solve using SVD
        elif self.method == "svd":
            Xb = np.hstack((np.ones((X.shape[0], 1)), X))
            beta = np.dot(np.linalg.pinv(Xb), y)
            self.b_ = np.array([beta[0]])
            self.w_ = beta[1:].reshape(X.shape[1], 1)

        return self

    def _normal_equation(self, X, y):
        """Solve linear regression analytically."""
        Xb = np.hstack((np.ones((X.shape[0], 1)), X))
        w = np.zeros(X.shape[1])
        z = np.linalg.inv(np.dot(Xb.T, Xb))
        params = np.dot(z, np.dot(Xb.T, y))
        b, w = np.array([params[0]]), params[1:].reshape(X.shape[1], 1)
        return b, w

    def _net_input(self, X):
        """Compute the linear net input."""
        return (np.dot(X, self.w_) + self.b_).flatten()

    def _predict(self, X):
        return self._net_input(X)

    def _sum_squared_error_cost(self, y, y_val):
        errors = y - y_val
        return (errors**2).sum() / 2.0
